/* =========================================================== */
/* EXmacro.c    ALU and shifter environments of the DLX        */
/* ----------------------------------------------------------- */
/*  last update: 9823nov SMM                                   */
/* =========================================================== */

#include <stdio.h>
#include <math.h>
#include "techno.h"
#include "macro.h"
#include "EXmacro.h"


/* ----------------------------------------------------------- */
/*     ALU Environment (width: var.n)                          */
/* ----------------------------------------------------------- */
/* --- Arithmetic unit AU  (with fast CLA adder) ------------- */
int cau() 
{ return (var.n +2) * cxor + cCLAf(var.n);
}

int dau()
{ return 3*dxor + dCLAf(var.n);
}

int dau2()
{ return 3*dxor + dCLAf(2);
}

/* --- Comparator -------------------------------------------- */
int ccomp() 
{ return czero(var.n) + 2*cinv + 4*cand + 2*cor;
}

int dcomp() 
{ return max(dinv+dand, max(dzero(var.n)+dor, dzero(var.n)+dinv))
          + dand + dor;
}

/* --- ALU --------------------------------------------------- */
int calu() 
{ return  cau() +ccomp() +var.n *(cand +cor +cxor) +5*var.n*cmux;
}

int dALU()
{ int delay;

  delay = max(dau() +dmux, max(dand, max(dor, dxor)) +3*dmux);
  return max(delay, dau()+dcomp()) + dmux;
}

/* --- ALU glue logic ---------------------------------------- */
int caluglue() 
{ return 2*3*cmux + cor;
}

int daluglue()                  /* delay of alu control */ 
{ return 2*dmux + dor;
}

/* --- ALU environment: data paths + glue -------------------- */
int cALUenv() 
{ return calu() + caluglue();
}

int dALUenv() 
{ return daluglue() + dALU();
}

int dALUe2() 
{ return daluglue() + dau2();
}

int dALUovf() 
{ return daluglue() + dau();
}

/* ---------------------------------------------------------- */
/*    Shifter Environments                                    */
/* ---------------------------------------------------------- */
/* --- Distance computation (Dist) -------------------------- */
int cdist() 
{ return cinc(5) + 5*cinv + 10*cmux;
}

int ddist() 
{ return dinv + dinc(5) + 2*dmux;
}

/* --- Fill Bit --------------------------------------------- */
int cfill()
{ return 2*cand;
}

int dfill()
{ return 2*dand;
}

/* --- Mask Generation (Mask) ------------------------------- */
int cmask() 
{ return chdec(5) + 2*var.n*cmux;
}

int dmask()
{ return dhdec(5) + 2*dmux;
} 

/* --- Environment SHenv ------------------------------------ */
int cshenv()
{ return  ccls(var.n) +cdist() +cfill() +cmask() +var.n * cmux;
}

int dshenv() 
{ return max(ddist()+dcls(var.n), max(dfill(), dmask())) + dmux;
}

/* --- Fill Bit for SH4L ------------------------------------ */
int cLfill()
{ return 5*cmux + cand + cinv;
}

int dLfill()
{ return max(dinv, 3*dmux) + dand;
}

/* --- Environment SH4L ------------------------------------- */
int cSH4L()
{ return  (3*var.n -8)*cmux + cnand + cLfill();
}

int dSH4L() 
{ return max(2*dmux, max(dLfill(), dnand)) + dmux;
}

/* --- Environment SH4Lenv (no SPR registers) --------------- */
int cSH4Le()
{ return  var.n*cmux + cSH4L();
}

int dSH4Le() 
{ return dSH4L() + dmux;
}

/* --- Environment SH4Lenv (with interrupts) ---------------- */
int cSH4Lie()
{ return  cSH4Le();
}

int dSH4Lie() 
{ return dSH4L() + dmux;
}

/* --- Environment SH4Lenv (with SPR + FPU)  ---------------- */
int cSH4Lif()
{ return  6*var.n*cmux + cSH4L();
}

int dSH4Lif() 
{ return max(dSH4L(), dmux) + 2*dmux;
}

