/***************************************************************
**  cache.c             k-way, sectored cache with interface  **
** ---------------------------------------------------------- **
**  990806 SMM 0.1                                            **
***************************************************************/

#include <math.h>
#include "techno.h"
#include "macro.h"
#include "hardctr.h"
#include "cache.h"

extern struct variable_parameters var;

struct cache_parameters cache = {0,0,0,0,1,0};
struct cacheIF_inputs cacheIF = {0,0,0};

/* ---------------------------------------------------------- */
/*     Cost of the cache and cache interface                  */
/* ---------------------------------------------------------- */
/* === Cost: direct mapped cache ============================ */
int Cch1()                
{ int cost, Line, Sector;

  Line = power2(cache.line);
  Sector = power2(cache.sector);

  cost = cram(Line,1) +cram(Line, cache.tag) +cEQ(cache.tag +1)
       + power2(cache.boff) * cram(Line * Sector, 8);

  return  cost;
}

/* === cost of an LRU cache replacement circuit ============= */
int Cactive()                
{ int k;

  k = cache.way;
  return  cdecf(ld(k)) + k*cmux;
}

int CHsel()                
{ return  2*cache.way * cmux * ld(cache.way);
}

int CLRUup()                
{ int k, cost;

  k = cache.way;
  if (k <= 2) 
    return 0;
  else
  { cost = CHsel() + cenc(ld(k));
    return  cost + k *(cEQ(ld(k)) + cand) + cPP(k) *cor;
  }
}

int CLRU()                
{ int k, K, l, cost;

  l = cache.line;
  k = cache.way;

  if (k <= 2)
    return  cram2(power2(l),1) +cff*(l+1) +cEQ(l) +2*cmux 
	    +cxnor +cinv;
  else
  { K =  k * ld(k);
    cost = cram2(power2(l), K) + K*(cff + 3*cmux);
    cost += cff*l + cEQ(l) + cor;
    return   cost + Cactive() + CLRUup();
  }
}

/* === cost : k-way set assoziative cache + LRU ============= */
int Cksel()                
{ int cost;

  cost = cache.way * cand + ctree(cache.way) * cor;
  return  cost * power2(cache.boff) *8 + ctree(cache.way) *cor ;
}

int CWadapt()                
{ return cache.way * (power2(cache.boff) +2) * cand;
}

int Cchk()                
{ return  cache.way *(cff + Cch1()) +Cksel() +CWadapt() +CLRU();
}


/* --- Cost: address generator (line fill) ------------------ */
int Cadrg() 
{ int cost, sec, Bytes;

  sec   = cache.sector;
  Bytes = power2(cache.boff);

  cost  = cff *sec + cEQ(sec) + cinc(sec); 
  return cost + 3*sec*cmux + Bytes *(cand + cor);
}

/* --- Cost: cache interface (with forwarding) -------------- */
int CchIF()
{ int cost, Bytes;

  Bytes = power2(cache.boff);
  cost = (cache.way <=1 ? Cch1() : Cchk());

  return cost + Cadrg() + Bytes *8 *(cff + 3* cmux);
}


/* ---------------------------------------------------------- */
/*     Delays of the cache and cache interface                */
/* ---------------------------------------------------------- */
/* === Delay: direct mapped cache =========================== */
int Dch1_Do()                                 /* -> Data, Do */
{ return dram( power2(cache.line + cache.sector), 8);
}

int Dch1_tag()                                 /* -> Dir, tag */
{ return dram( power2(cache.line), cache.tag);
}

int Dch1_v()                                     /* -> valid  */
{ return dram( power2(cache.line), 1);
}

int Dch1_hit()                                /* cache -> hit */
{ return Dch1_tag() + dEQ(cache.tag +1);
}

/* === Delay: LRU replacement circuit ======================= */
int Dway_ev()                           /* ev -> active way */
{ return  ddecf(ld(cache.way)) + dmux;
}

int Dway_hit()                           /* hit -> active way */
{ return  dmux;
}

int DHsel()                
{ return  2* dmux;
}

int DLRU_hit()                
{ if (cache.way <= 2)
   return dmux + dxnor + dff;
  else
   return  dand + dmux + dPP(cache.way) + DHsel() + dff;
}

int DLRU_h()                
{ int k;

  k = cache.way;
  if (k <= 2)
   return dmux + dxnor + dff;
  else
   return  denc(ld(k)) + dEQ(ld(k)) +DLRU_hit();
}

int DLRU_a()                
{ int delay, k, K, l;

  l = cache.line;
  k = cache.way;

  if (k <= 2)
   return  max(dram2(power2(l),1), dEQ(l)) +2*dmux +dxnor +dff;
  else
  { K = k * ld(k);
    delay = max(Dway_ev() +dff, dEQ(ld(k)) + DLRU_hit());
    return  delay + dmux + max(dEQ(l), dram2(power2(l),K));
  }
}

/* === Delay: LRU replacement circuit ======================= */
int DWadapt()                
{ return (cache.way < 2 ? 0 : dand);
}

int Dksel()                
{ return (cache.way < 2 ? 0 : dand + dtree(cache.way)*dor);
}

/* --- Delay of data Do and of the hit signal --------------- */
int Dch_Do()                
{ return Dch1_Do() + Dksel();
}

int Dch_hit()                
{ return Dch1_hit() + Dksel();
}

/* --- Delay: history update, due to k-way cache ------------ */
int Dch_1_H()                
{ int delay;

  delay = Dch1_hit() + DLRU_h();
  return (cache.way < 2 ? 0 : delay);
}

int Dch_H()                
{ int delay;

  delay = max(Dch_1_H(), DLRU_a());
  delay = max(delay, Dch_hit() + DLRU_hit());
  return (cache.way < 2 ? 0 : delay);
}

/* --- Delay: input -> storage of k-way cache --------------- */
int Dcha_ch()                
{ return  max(Dch_H(), max(Dch1_tag(), Dch1_Do()));
}

int Dchc_ch()                      /* any control signal */
{ return  max(Dch_H(), Dch1_tag() + DWadapt());
}

int DchDi_ch()                
{ return  Dch1_Do();
}

/* --- Delay: address generator (line fill) ----------------- */
int Aadr_a()                              /* Adrgen -> ma, ca */
{ return  max(cacheIF.A, cacheIF.csig) + dmux;
}

int Aadr_rs()                             /* Adrgen -> rs */
{ return  cacheIF.A + dEQ(cache.sector);
}

int Aadr_CDw()                             /* Adrgen -> CDw */
{ return  cacheIF.csig + dand + dor;
}

int Tadr()                              /* Cycle time Adrgen */
{ return  max(cacheIF.csig, dinc(cache.sector)) 
	   + dmux + dff +delta;
}

/* --- Delay: cache interface (with forwarding) ------------- */
int AchIF_hit()                             /* cacheIF -> hit */
{ return  Aadr_a() +Dch_hit();
}

int AchIF_Do()                              /* cacheIF -> Do */
{ return  Aadr_a() +Dch_Do();
}

int AchIF_Dout()                          /* cacheIF -> Dout */
{ return max(Aadr_rs(), AchIF_Do()) + 2*dmux;
}

int AchIF_Di()                              /* cacheIF -> Di */
{ return max(cacheIF.Din, cacheIF.csig) + dmux;
}

int AchIF_MAdr()                          /* cacheIF -> MAdr */
{ return  Aadr_a();
}

int TchIF()                            /* cycle time CacheIF */
{ int delay, delay1;

  delay = max(Tadr(), Aadr_CDw() + Dchc_ch() +delta);
  delay1 = max(Aadr_a() + Dcha_ch(), AchIF_Di() + DchDi_ch());
  return  max(delay, delay1 + delta);
}

int DchIFM_Dout()                          /* MDat -> Dout */
{ return 2*dmux;
}

int DchIFM_ch()                          /* MDat -> cacheIF */
{ return dmux + DchDi_ch();
}

