/***************************************************************
**  evaluate.c      quality ratio Q(q,A,B) and equal quality  **
**                  parameter EQ(A,B) of two designs A,B      **
**  950801 TGR 0.1                                            **
**  950811 SMM 0.2  (update of EQ(), degenerated cases)       **
***************************************************************/
/* q      : quality parameter q                               */
/* ca, ta : cost, dealy of A      cb, tb : cost, dealy of B   */
/* ---------------------------------------------------------- */
/* Let F be the faster and S be the slower of the designs.    */
/* EQ(S,F) defined for                                        */
/*   a) ts > tf  and  cs <= cf;   b) ts >= tf  and  cs < cf   */
/* Degenerated cases:                                         */
/*  ts = tf and cs = cf: always same quality:  EQ(S,F) := +2  */
/*  ts > tf and cs > cf: never same quality:   EQ(S,F) := -2  */
/* ---------------------------------------------------------- */

#include <math.h>

double eval_Q(q,ca,ta,cb,tb) 
double q;
int ca,ta,cb,tb;
{ double c,t,r;
  
  r = (1.0*cb)/(1.0*ca);
  c = pow(r, q);
  r = (1.0*tb)/(1.0*ta);
  t = pow(r, 1.0-q);
  r = c*t; 
  return r;
}

double eval_EQ(ca,ta,cb,tb) 
int ca,ta,cb,tb;
{ double c,t,r;
  
  c = (1.0*ca) - (1.0*cb);
  t = (1.0*ta) - (1.0*tb);
  r = c * t;

  if ( r > 0)
  { /* faster design is cheaper */
    return -2;
  }
  else 
  { if ((c == 0) && (t == 0))
    { /* both designs of equal quality */
      return +2;
    }
    else
    { /* standard case */
      c = (1.0*ca)/(1.0*cb);
      t = (1.0*ta)/(1.0*tb);
      r = log(t)/log(t/c);
      return r;
    }
  }
}

