/* =========================================================== */
/* fpround.c    architectural module   rounding-unit           */
/* ----------------------------------------------------------- */

#include <stdio.h>
#include <math.h>
#include "techno.h"
#include "macro.h"
#include "hardctr.h"
#include "evaluate.h"

struct variable_parameters {
  int aRM;                /* acc. delay of rounding mode RM */
  int aEN;                /* acc. delay of UNFen and OVFen */
} Fvar;

extern double romfac;
int  FASTINC;

/*=============================================================*/
/*  Floating Point Unpacking  FPunp                            */
/*=============================================================*/
/* --- Special Cases SpecUnp --------------------------------- */
int cSpecUnp()
{ return (4*cand + cnor);
}

int dSpecUnp()
{ return (dnor + dnand);
}

/* ----- circuit Unpack -------------------------------------- */
int cUnpack()
{ return(2*czero(11) + 2*czero(8) + czero(52) + clzero(53) 
	+ cincf(11) + cincf(8) + ccls(53) + 22*cinv + cmux*13
	+ cmux*52 + cmux*53 + 2*cor);
}

int aUnpack_reg()
{ return dzero(11) +dmux +dinv +
     max(dincf(11)+dor, dlzero(53)+dcls(53)+dmux);
}

int aUnpack_flag()
{ return dmux + max(dzero(11)+dinv, dzero(52));
}

/* --- NaNselect (NaNsel) ------------------------------------ */
int cNaNsel()
{ return cmux*53;
}

int dNaNsel()
{ return dmux;
}

/* --- circuit FPunp (toplevel) ------------------------------ */
int cFPunp()
{ return 2*cUnpack() +2*cSpecUnp() +cNaNsel() +203*cff;
}

int aFPunp_flag()
{ return aUnpack_flag() +dSpecUnp();
}

int aFPunp()
{ return max(aFPunp_flag() +dNaNsel(), aUnpack_reg());
}


/*=============================================================*/
/*    Floating Point Condition Test  FCon                      */
/*=============================================================*/
/* --- circuit FPtest ---------------------------------------- */
int cFPtest()
{ return 13*cand +8*cor +3*cinv +cnor +cnand;
}

int dFPtest()
{ return dinv +4*dand +3*dor;
}

/* --- circuit FCon (toplevel) ------------------------------- */
int cFCon()
{ return cCLAf(64) +cEQ(64) +63*cinv + cFPtest() +cand +cnor;
}

int aFCon()
{ int delay;

  delay = max(dEQ(64), dinv +dCLAf(64));
  return  max(delay, aFPunp_flag()) +dFPtest() +dand;
}


/*=============================================================*/
/*    FP Add/Sub circuit                                       */
/*=============================================================*/
/* ---  Special Cases  SpecAS                            ----- */
int cSpecAS()
{ return  5*cor +cxor +cinv +3*cand;
}

int dSpecAS()
{ return  dxor +2*dor +dinv +2*dand;
}

/* --- Sign computation: SignSelect -------------------------- */
int cSignSelAS()
{ return  2*cand +cnor +4*cmux;
}

int dSignSelAS()
{ return  dand + max(3*dmux, dnor);
}

/*-------------------------------------------------------------*/
/*    Significand Add                                          */
/*-------------------------------------------------------------*/
/* --- Absolute Value: Abs ----------------------------------- */
int cAbs(n)
int n;
{ if (FASTINC < 2)
   return((n-1)*cinv + cincf(n-1) + cmux*(n-1));
  else
   return((n-1)*cinv + cCSI(n-1) + cmux*(n-1));
}

int dAbs(n)
int n;
{ if (FASTINC < 2)
   return(dinv + dincf(n-1) + dmux);
  else
   return(dinv + dCSI(n-1) + dmux);
}

/* --- Sign -------------------------------------------------- */
int cSign()
{ return  2*cand +cnand +cor;
}

int dSign()
{ return  dand +dor +dnand;
}

/*--- SigAdd --------------------------------------------------*/
int cSigAdd()
{ return  58*cxor +cCLAf(58) +cSign() +cAbs(58) +czero(58);
}

int dSigAdd_flag()
{ return  dxor +dCLAf(58) +max(dzero(58), dSign());
}

int dSigAdd_fs()
{ return  dxor +dCLAf(58) +dAbs(58); 
}

/*-------------------------------------------------------------*/
/*  Alginment Shifter AlignShift                               */
/*-------------------------------------------------------------*/
/* --- Stick bit comp. Sticky -------------------------------- */
int cSticky()
{ return  chdec(6) +55*cand +cor*ctree(55);
}

int dSticky()
{ return  dhdec(6) +dand +dor*dtree(55);
}

/* ---  Swap ------------------------------------------------- */
int cSwap()
{ return  cmux*54 +cmux*55;
}

int dSwap()
{ return dmux;
}

/* --- Limit ------------------------------------------------- */
int cLimit()
{ return  11*cinv +cmux*11 +cor*ctree(7) +6*cor;
}

int dLimit()
{ return (dinv + dmux + dor*dtree(7) + dor);
}

/* --- Exponent subtract: ExpSub ----------------------------- */
int cExpSub()
{ return  cCLAf(12) +11*cinv;
}

int dExpSub()
{ return dinv+ dCLAf(12);
}

/* --- Alignment shift: AlignShift (toplevel) ---------------- */
int cAlignShift()
{ return  cExpSub() +cLimit() +cSwap() +ccls(55) 
          +cSticky() +cxor +cmux*11;
}

int dAlignShift()
{ return dExpSub() + max(dLimit(), dSwap()) 
       + max(dcls(55), dSticky());
}

/*=== FP Add/Sub ==============================================*/
int cAddSub()
{ return  cAlignShift() +cSigAdd() +cSignSelAS() 
          +cxor +182*cff +cSpecAS();
}

int tAddSub()
{ return  dxor +max(dAlignShift(), dSpecAS()) +dff +delta;
}

int aAddSub()
{ return  max(dSigAdd_fs(), 
	      max(Fvar.aRM, dSigAdd_flag()) +dSignSelAS());
}


/*=============================================================*/
/*  FP Multiply/Divide: MulDiv                                 */
/*=============================================================*/
/* --- Special cases SpecMD ---------------------------------- */
int cSpecMD()
{ return  12*cor +10*cand +cnor +cinv +3*cmux;
}

int dSpecMD()
{ return  dinv +2*dand +4*dor +2*dmux;
}

/* --- Sign/Exponent: SignExp -------------------------------- */
int cSignExpMD()
{ return  cxor +cinv*23 +cmux*13 +cmux*11 +cA4_2(13) +cCLAf(13);
}

int aSignExpMD()
{ return (dinv + dmux + dA4_2(13) + dCLAf(13));
}

/* --- Select Fd --------------------------------------------- */
int cSelectFd()
{ return  cincf(55) +cmux*(29+56+1+87) +cA3_2(116) +cCLAf(117) 
     +czero(117) +203*cinv +cand;
}

int dSelectFd()
{ return 2*dmux +max(dincf(55)+dmux,
	  2*dinv +dA3_2(116) +dCLAf(117) +dzero(117));
}

/*-------------------------------------------------------------*/
/*    Wallace Mul-Array with carry-save result (n,m)           */
/*                                                             */
/*   D = Dand + 2(r+1) dFA      \mu = r = ld(m)-2              */
/*   C = n*m*Cand + C4_2tree(n,m)                              */
/*   C4_2tree(n,m) = cFA*(n*m -2n +r*m)                        */
/*-------------------------------------------------------------*/
int cWallace(n,m)
int n,m;
{ int r, c42tree;

  r = ld(m) -2;
  c42tree = cFA()* (n*m -2*n +r*m);
  return n*m* cand + c42tree;
}

int dWallace(n,m)
int n,m;
{ int r;

  r = ld(m) -2;
  return dand + 2*(r+1)*dFA();
}

/*--- significand MulDiv SigfMD -------------------------------*/
int cSigfMD()
{ return  crom(256,8) +cWallace(58,58) +cCLAf(116) +cor*ctree(60) 
    +cSelectFd() +cmux*(58+57) +6*cdriv*58 +3*cff*116 +5*cff*58 
    +cinv*58 +cand*29;
}

int aSigfMD()
{ return  max(dSelectFd(), dCLAf(116) +dor*dtree(60)) +dmux;
}

int tSigfMD()
{ int delay;

  delay = max(ddriv +drom(256,8), dCLAf(116)) + dmux;
  return  max(delay, ddriv +dWallace(58,58)) + dff +delta;
}

/*==== FP Mul/Div unit (toplevel) =============================*/
int cMulDiv()
{ return cSigfMD() +cSignExpMD() +cSpecMD() +cff*72;
}

int aMulDiv()
{ return aSigfMD();
}

int tMulDiv()
{ int delay;

  delay = max(dSpecMD(), aSignExpMD()) +dff +delta;
  return  max(delta, tSigfMD());
}


/*=============================================================*/
/*   Floating Point rounder FPrnd                              */
/*=============================================================*/
/* ---   Special Cases  SpecFPrnd                        ----- */

int cSpecFPrnd()
{ return  cmux*(11 +2*52 +1 +64) +cinv*3 +5*cand +4*cor;
}

int  dSpecFPrnd()
{ return  2*dor +max(2*dmux, 2*dand +dinv);
}

/*--- Exponent Round ----------------------------------------- */
int cExpRnd()
{ return cxnor +cmux +2*cmux*63 +cinv +2*cand;
}

int dExpRnd()
{ return 3*dmux +dxnor;
}

/*--- Adjust exponent -----------------------------------------*/
int cAdjustExp()
{ return  cand*ctree(11) +3*cand +cinv +cmux*14;
}

int dAdjustExp()
{ return (dmux + dand*dtree(11) + dand + dmux);
}

/*--- Post Normalization PostNorm------------------------------*/
int cPostNorm()
{ return  cor +11*cmux;
}

int dPostNorm()
{ return max(dor, dmux);
}

/*-------------------------------------------------------------*/
/*    Significand Rounding SigRnd                              */
/*-------------------------------------------------------------*/
/* ---   Rounding Decision                               ----- */

int cRndDcs()
{ return (3*cand + 2*cor+ cmux + cxnor);
}

int dRndDcs()
{ return  dxnor + dand + dmux;
}

/* ---   SigRnd                                          ----- */
int cSigRnd()
{ if ((FASTINC == 0)|| (FASTINC ==2))
   return  cincf(53)+ cRndDcs() +3*cmux +cmux*53 +cmux*54 +cor;
  else
   return  cCSI(53)+ cRndDcs() +3*cmux +cmux*53 +cmux*54 +cor;
}

int aSigRnd()
{ if ((FASTINC == 0)|| (FASTINC ==2))
    return  max(dincf(53), Fvar.aRM +dRndDcs()) +2*dmux;
  else
    return  max(dCSI(53), Fvar.aRM +dRndDcs()) +2*dmux;
}

/* ---   p-Representative REPp ------------------------------- */
int cREPp()
{ return  cor*(ctree(29) +ctree(74) +1) +30*cmux;
}

int dREPp()
{ return  dor*dtree(74) +dor +dmux;
}

/*-------------------------------------------------------------*/
/*  Normalize Significand: NormShift                           */
/*-------------------------------------------------------------*/
/* ---  Flags (Tiny, OVF1, lz) ------------------------------- */
int cFlags()
{ int TINY, OVF;

  TINY = clzero(64) +cinv +cCLAf(13);
  OVF  = 2*cinv +5*cor +3*cand +cEQ(13);

  return TINY +OVF;
}

int aFlags()
{ int TINY, OVF;

  TINY = dlzero(64) +dinv +dCLAf(13);
  OVF  = dor +dand +max(3*dor+dand, dEQ(13));

  return max(TINY, OVF);
}

/*--- exponent normalization shift ExpNorm --------------------*/
int cExpNorm()
{ return  cA3_2(11)+ cA2(11) +cmux*(4+5+22) +3*cand +7*cinv;
}

int aExpNorm()
{ return  max(aFlags(), Fvar.aEN) +dand 
	   +4*dmux +dA3_2(11) +dA2(11);
}

/* --- shiht distance ---------------------------------------- */
int cShiftDist()
{ return  cCLAf(13) +cmux*13 +cinv +cand;
}

int aShiftDist()
{ return max(dCLAf(13), max(aFlags(), Fvar.aEN)+dinv+dand) +dmux;
}

/*---- Significand normalization shift   SigNormShift -------- */
int  cSigNormShift()
{ int cMASK;

  cMASK = 12*cinv +cmux*(12+6+64) +cor*ctree(6) +chdec(6)
	  +64*(cand+cinv);

  return  ccls(64) +cMASK + 2*64*cand;
}

int  dSigNormShift()
{ int dMASK;

  dMASK = dinv +3*cmux +dor*dtree(6) +dhdec(6) +dand;
  return  max(dcls(64), dMASK) + dand;
}


/*==== Normalization shift NormShift (toplevel) ===============*/
int cNormShift()
{ return  cFlags() +cExpNorm() +cShiftDist() +cSigNormShift();
}

int aNormShift()
{ return  max(aShiftDist()+dSigNormShift(), aExpNorm());
}

/*==== Floating Point Rounder (toplevel) ======================*/
int cFPrnd()
{ return  cNormShift() +cREPp() +cSigRnd() +cPostNorm()
     +cAdjustExp() +cExpRnd() +cSpecFPrnd() +cff*140;
}

int aFPrnd()
{ return  aSigRnd() +dPostNorm() +dAdjustExp() +dExpRnd() 
          +dSpecFPrnd();
}

int tFPrnd()
{ return  aNormShift() +dREPp() +dff +delta;
}


/*=============================================================*/
/*    Fixed Point Unpacker FXunp                               */
/*=============================================================*/
int cFXunp()
{ return  cinv*32 +cincf(32) +czero(32) +cmux*32;
}

int aFXunp()
{ return max(dinv +dincf(32)+dmux, dzero(32));
}

/*=============================================================*/
/*    Conversion Cvt                                           */
/*=============================================================*/
int cCvt()
{ return  cor;
}

int dCvt()
{ return dor;
}


/*=============================================================*/
/*   Fixed Point Rounder FXrnd                                 */
/*=============================================================*/
/* ---   Special Cases  SpecFX                           ----- */

int cSpecFX()
{ return  32*cxor +2*cxor +cand + cnor +5*cor; 
}

int  dSpecFX()
{ return  dor +max(dxor+dor, dand +dor);
}

/*-------------------------------------------------------------*/
/*    Significand Rounding SigRndX                             */
/*-------------------------------------------------------------*/
/* ---   Rounding Decision  RndDcs                       ----- */
int cSigRndX()
{ return  cincf(33)+ cRndDcs() +cmux*33 +2*cxor +cor;
}

int aSigRndX()
{ return  max(dincf(33), Fvar.aRM +dRndDcs()) +dmux +dxor;
}

/* ---   p-Representative REPpX ------------------------------ */
int cREPpX()
{ return  cor*ctree(95);
}

int dREPpX()
{ return  dor*dtree(95);
}

/*-------------------------------------------------------------*/
/*  Normalize Significand: NormShiftX                          */
/*-------------------------------------------------------------*/
/*---- Significand normalization shift   SigNormShift -------- */

/* --- shiht distance ShiftDistFX ---------------------------- */
int cShiftDistX()
{ return  cinv*6 +cCLAf(13) +cA3_2(13);
}

int dShiftDistX()
{ return  dA3_2(13) +dCLAf(13) +dinv;
}

/* ---  FXflags ---------------------------------------------- */
int cFXflags()
{ return cor*ctree(5) +cand +cinv;
}

int dFXflags()
{ return dor*dtree(5) +dand;
}

/*==== Normalization shift NormShiftX (toplevel) ==============*/
int cNormShiftX()
{ return  cFXflags() +cShiftDistX() +cSigNormShift();
}

int dNormShiftX()
{ return  max(dShiftDistX()+dSigNormShift(), dFXflags());
}

/*==== Fixed Point Rounder FXrnd (toplevel) ===================*/
int cFXrnd()
{ return  cNormShiftX() +cREPpX() +cSigRndX() +cSpecFX() +cff*40;
}

int aFXrnd()
{ return  aSigRndX() +dSpecFX();
}

int tFXrnd()
{ return  dNormShiftX() +dREPpX() +dff +delta;
}

/*=============================================================*/
/*    Floating Point Unit FPU                                  */
/*=============================================================*/
int cFPUreg()
{ return  129*(cff +6*cdriv);
}

int cFPU()
{ return  cFCon() +cFPunp() +cFXunp() +cCvt() +cMulDiv()
	  +cAddSub() +cFPrnd() +cFXrnd()+ cFPUreg();
}

int aFPU()
{ return max(aFCon(), max(aFXrnd(), aFPrnd()));
}

int aFPU_fr()
{ int delay;

  delay = max(aFPunp() +dCvt(), aFXunp()); 
  return  max(delay, max(aMulDiv(), aAddSub())) +ddriv; 
}

int tFPU()
{ int delay;

  delay  = max(aFPU_fr() +dff+delta, tFXrnd());
  delay  = max(delay, tFPrnd());
  return   max(delay, max(tMulDiv(), tAddSub()));
}


/* === Print functions ====================================== */
void print_line()
{
  printf(
  "---------------------------------------------------------\n");
  return;
}

void print_head(str)
char *str;
{ printf("\n"); 
  print_line();
  printf("\t\t %30s\n",str); 
  printf("%21s:     none     RND    Abs   RND/Abs\n","CSI incrementer"); 
  print_line();
  return;
}

void print_value(str,func)
char *str;
int (*func)();
{ printf("%21s:",str); 
  FASTINC=0;  printf("\t%7d",(*func)());
  FASTINC=1;  printf("\t%7d",(*func)());
  FASTINC=2;  printf("\t%7d",(*func)());
  FASTINC=3;  printf("\t%7d\n",(*func)());
  return;
}


/* ---------------------------------------------------------- */
/*     Evaluation of the design                               */
/* ---------------------------------------------------------- */
int main()
{
  int i;
  char my_str[200];             /* temporary strings */
  int cost, delay;

  init_techno(MOTOROLA);        /* select technology */
  romfac = 8.0;

  Fvar.aRM = Fvar.aEN = 0;

/* ---------------------------------------------------------- */
/*    Cost of the FPU                                         */
/* ---------------------------------------------------------- */

  print_head("Kosten");
  print_value("FPU",cFPU);
  print_line();            
  print_value("cFPunp",cFPunp);
  print_value("cFXunp",cFXunp);
  print_value("cFCon",cFCon);
  print_value("cAddSub",cAddSub);
  print_value("cMulDiv",cMulDiv);
  print_value("cCvt",cCvt);
  print_value("cFPrnd",cFPrnd);
  print_value("cFXrnd",cFXrnd);
  print_value("cFPUreg",cFPUreg);
  print_line();

  printf("Kosten: Mul/Div\n");
  print_line(); 
  print_value("cSigfMD",cSigfMD);
  print_value("cSignExpMD",cSignExpMD);
  print_value("cSpecMD",cSpecMD);
  print_line(); 
  printf("Kosten SigfMD:  \t romfac = %3.1f \n", romfac);
  cost = crom(256,8); printf("\t ROM = %d,",cost);
  cost = cWallace(58,58); printf("\t Wallace = %d\n",cost);
  cost = cCLAf(116); printf("\t CLA = %d,",cost);
  cost = cSelectFd(); printf("\t Select = %d\n",cost);
  print_line();

/* ---------------------------------------------------------- */
/*   Delay and cycle time                                     */
/* ---------------------------------------------------------- */
  for(i=0; i<2; i++)
  { Fvar.aRM = Fvar.aEN = 100*i;
    printf("aRM = aEN = %d\n",Fvar.aRM);

    print_head("Acc.Delay Output");
    print_value("aFPU",aFPU);
    print_line(); 
    print_value("aFCon",aFCon);
    print_value("aFXrnd",aFXrnd);
    print_value("aFPrnd",aFPrnd);
    print_line(); 

    print_head("Acc.Delay Bus FR");
    print_value("aFPU_fr",aFPU_fr);
    print_line(); 
    print_value("aFPunp",aFPunp);
    print_value("+dCvt",dCvt);
    print_value("aFXunp",aFXunp);
    print_value("aMulDiv",aMulDiv);
    print_value("aAddSub",aAddSub);
    print_value("dSigAdd(fl)",dSigAdd_flag);
    print_value("dSigAdd(fs)",dSigAdd_fs);
    print_line(); 
  
    print_head("Zykluszeit");
    print_value("tFPU",tFPU);
    print_line(); 
    print_value("tFXrnd",tFXrnd);
    print_value("tFPrnd",tFPrnd);
    print_value("tAddSub",tAddSub);
    print_value("tMulDiv",tMulDiv);
    print_line(); 

    print_head("tFPrnd:");
    print_value("aNormShift",aNormShift);
    print_value("aShiftDist",aShiftDist);
    print_value("dSigNormShift",dSigNormShift);
    print_value("aFlags",aFlags);
    print_value("aExpNorm",aExpNorm);
    print_line(); 

    print_head("aFPrnd:");
    print_value("aSigRnd",aSigRnd);
    print_value("dPostNorm",dPostNorm);
    print_value("dAdjustExp",dAdjustExp);
    print_value("dExpRnd",dExpRnd);
    print_value("dSpecFPrnd",dSpecFPrnd);
    print_line(); 
  }

return(0);
}
