/***************************************************************
**  hardctr.h      hardwired control logic module             **
**   Mealy and Moore implementations of the control automaton **
** ---------------------------------------------------------- **
**  950801 TGR 0.1 (from SMM dlx1.c)                          **
**  950803 SMM 0.2 (changed names lmax, lsum)                 **
**  950805 SMM 0.3 (added functions aILL,cILL)                **
**  950809 SMM 0.4 (added print functions)                    **
**  960417 SMM 0.5 (Mealy automaton)                          **
***************************************************************/
/*  Accumulated delay of the inputs of the automaton:         */
/*  (to be defined in the architecture file)                  */
/*                                                            */
/*  int aIncl() : Delay of the clear signal                   */
/*------------------------------------------------------------*/

#ifndef CONTROL_H
#define CONTROL_H

extern int set_contype();    /* set ctrl automaton type */
                             /* par: precomputed, binary */
extern int set_conpar();     /* set  automaton parameters */
                             /* states, inputs, outputs,
                                #monomials, length max/sum,
                                nu max/sum, fanin max/sum */
extern int aOut();           /* accum. delay of outputs */
extern int aNSE();           /* accum. delay of next state */
extern int cauto();          /* cost of the automaton */
extern int tauto();          /* cycle time of the automaton */

                             /* circuit ILL detects illegal
                                 instructions */
extern int cILL();           /* cost of ILL */
extern int aILL();           /* accumulated dealy of ILL */

extern void forall_line();     /* print horizontal line */
extern void forall_header();   /* print two line header */
extern void forall_print();    /* evaluate function and print */

struct control_parameters {
  int states, zeta;            /* #States, ld(states) */
  int sigma, gamma;            /* #input/output signals */
  int monom;                   /* #monomials m in state
                                   transitions */
  int lmax, lsum;              /* max, sum of literals in m */
  int numax, nusum;            /* max, sum of signal
                                    activations */
  int faninmax, faninsum;      /* max, sum fanin of states */
  int type;                    /* moore/pmoore/mealy/pmealy */
  int binary;                  /* unary/binary */
  int aIN;                     /* acc. delay of inputs */
  int aclr;                    /* acc. delay: clear signal */
};

extern struct control_parameters con;

#endif CONTROL_H


