/***************************************************************
**  macro.h        commonly used functional modules           **
**  950801 TGR 0.1 (from SMM dlx1.c)                          **
**  950803 SMM 0.2 (adding left shifter)                      **
**  950812 SMM 0.3 (adding full adder FA)                     **
***************************************************************/

#ifndef MACRO_H
#define MACRO_H

extern int max();      /* max of two integers */
extern int ld();       /* log base 2, rounded to next integer */
extern int power2();   /* 2^n */

extern int cdecs(), ddecs();    /* decoder, slow */
extern int cdecf(), ddecf();    /* decoder, fast */
extern int chdec(), dhdec();    /* half decoder */
extern int cenf(),  denf();     /* encoder, with flag */
extern int cenc(),  denc();     /* encoder, no flag */

extern int ctree(), dtree();    /* binary tree, #gates, depth */
extern int czero(), dzero();    /* zero tester */
extern int dfzero();            /* fast zero tester */
extern int clzero(), dlzero();  /* leading 0 counter */
extern int cEQ(),   dEQ();      /* equality tester */
extern int cinc(),  dinc();     /* incrementer, ripple carry */
extern int cincf(), dincf();    /* incrementer, cond. ripple */
extern int cCSI(),  dCSI();     /* Cond Sum Incrementer */ 
extern int ccls(),  dcls();     /* cyclic left shifter */ 

extern int cFA(),   dFA();      /* full adder */
extern int cRCA(),  dRCA();     /* adder (carry ripple) */
extern int cPP(),   dPP();      /* parallel prefix */
extern int cCLA(),  dCLA();     /* adder (carry lookahead) */
extern int cCLAf(), dCLAf();    /* fast carry lookahead adder  */
extern int cCSA(),  dCSA();     /* adder (conditional sum) */
extern int cA2(),   dA2();      /* A2 adder */
extern int cA3_2(), dA3_2();    /* 3/2 adder */
extern int cA4_2(), dA4_2();    /* 4/2 adder */

extern int ckmux(), dkmux();    /* mux: k inputs -> 1 output */
				/* inputs: (n=width, k=number) */

extern int cFL1b(), dFL1b();    /* find last 1: result in binary */
extern int cFL1u(), dFL1u();    /* find last 1: clear all other bits */
extern int cFF0(), dFF0();    /* find first 0: result in binary */

#endif

