/* =========================================================== */
/* pdlx.c       architectural module   pipelined DLX core      */
/*         no interrupt handling, hardwired control, CLA adder */
/* ----------------------------------------------------------- */
/*  last update: 99/12/29 SMM                                  */
/* =========================================================== */

#include <stdio.h>
#include <math.h>
#include "techno.h"
#include "macro.h"
#include "EXmacro.h"
#include "hardctr.h"
#include "evaluate.h"

struct variable_parameters var;

int DLXp;
/* ----------------------------------------------------------- */
/* DLXp = 0: no forwarding, no stalling                        */
/*        1: with forwarding, no stalling                      */
/*        2: with forwarding, with stalling                    */
/* ----------------------------------------------------------- */
   
int stage;
/* ----------------------------------------------------------- */
/* identifies the two control automata                         */
/* stage = 1: Mealy       2: Moore                             */
/* ----------------------------------------------------------- */
/* parameter sets                                              */
/* states, zeta - in, out - monom, lmax, lsum - numax, nusum - */
/* fanmax, fansum - type, binary - aIN, aclr                   */
/* ----------------------------------------------------------- */
struct control_parameters conMealy =
       {1,0, 0,8, 5,10,22, 2,11, 1,0, 2,0, 0,0};
struct control_parameters conMoore = 
       {10,4, 12,11, 15,11,104, 9,39, 3,15, 1,0, 0,0};

/*=============================================================*/
/*   Cost of the data paths and its environments               */
/*=============================================================*/
/*---- IR Environment ----------------------------------------*/
int cIRenv()
{ return 15*cmux +var.n*cff;
}

/*---- CAddr (destination address) ---------------------------*/
int cCAddr()
{ return 2* ld(var.A)* cmux;
}

/*---- GPR register file -------------------------------------*/
int cGPR()
{ return  cram2(var.A, var.n) 
	  + 2* (czero(ld(var.A)) + cinv + var.n* cand);
}

/*---- execute environment EXenv -----------------------------*/
int cEXenv()
{ return var.n* (2*cmux + 3*cdriv) + cALUenv() + cshenv();
}

/*-------------------------------------------------------------*/
/*   PC environment (delayed PC)                               */
/*-------------------------------------------------------------*/
/*---- PC glue logic -----------------------------------------*/
int cPCglue()
{ return 2* cor + cand + cxor + czero(var.n);
}

/*---- PC data paths -----------------------------------------*/
int cPCenv()
{ return var.n* (2*cff + 4*cmux) + cCLAf(var.n) 
	  + cincf(var.n-2) + cPCglue();
}

/*---- forwarding data paths (+control, 2x Forw3) ------------*/
int cforw()
{ int cost, ldA;

  ldA = ld(var.A);
  cost = ldA*cxor + czero(ldA) + 2*cand + var.n*cmux;
  if (DLXp == 2) cost+= cand;

  cost = 3*cost + 3*cand + 2*cinv + ctree(ldA)*cor;
  return (DLXp == 0) ? 0 : 2*cost;
}

/*-------------------------------------------------------------*/
/*   whole DLX data paths                                      */
/*-------------------------------------------------------------*/
int cBuffer()
{ return 3*ld(var.A)*cff + 3*12*cff;
}

/*=============================================================*/
int cDP()
{ return cPCenv() + cIRenv() + cEXenv() + cGPR() + cSH4Le()
	 + cCAddr() + var.n *7*cff + cBuffer()
	 + cforw();
}

/*=============================================================*/
/*   Cost of the control: stall, automata, IMC, DMC            */
/*=============================================================*/
int cIMC()
{ return 0;
}

int cDMC()
{ return cdecf(2) + 12*cand + 5*cor + 2*cinv;
}

int cstall()
{ int cdhaz, cost;

  cdhaz = 4*cand + 2*cinv + 3*cor;
  cost = 15*cand + 2*cnor +cor + 3*cff + cinv;

  if (DLXp <= 1)
    return cost;
  else
    return cost + cdhaz + cinv + cor + 3*cand;
}

/* --- control automata -------------------------------------- */
int cpauto()
{ int cost;

  con = conMealy; stage = 1; cost = cauto();   /* Mealy */
  con = conMoore; stage = 2; cost += cauto();  /* pMoore(EX-WB) */
  return cost;
}

/* ------------------------------------------------------------ */
/* signal of type y/z are just buffered  (5 FFs)                */
/* valid signals v[4:2] are used for forwarding/stalling        */
/* ------------------------------------------------------------ */
int cCONbuffer()
{ int cvalid;

  cvalid = 6*cff + cinv;      /* valid signals on forwarding */
  return (DLXp == 0) ? 5*cff : 5*cff+cvalid;
}


int cCON()
{ return cIMC() +cDMC() +cstall() +cpauto() +cCONbuffer();
}


/*=============================================================*/
/*   Control Automata: Delay of the control signals            */
/*=============================================================*/
int acsMealy()                 /* Mealy signals: ID stage */
{ con = conMealy;
  return aOut();
}

int acsMoore()                  /* Moore signals */
{ con = conMoore;
  return aOut();
}

/*=============================================================*/
/*   Delay / Cycle Time of the data paths environments         */
/*=============================================================*/
/* --- execute stage EXenv ----------------------------------- */
int aEX()
{ return acsMoore() +dmux +max(dALUenv(), dALUe2()+dshenv())
	      + ddriv;
}

int tEX()
{ return aEX() + dff + delta;
}

/* ---------------------------------------------------------- */
/*  WB stage: SH4L GPR (# Registers: var.A, width: var.n)     */
/* ---------------------------------------------------------- */
int dGPRread()                   
{ return max(dram2(var.A, var.n), dzero(ld(var.A)) +dinv) 
	 + dand;
}

int dGPRwrite()                   
{ return dram2(var.A, var.n);
}

int aC_4()                  /* input Cdata of GPR */
{ return acsMoore() + dSH4Le();
}

int tWB()                   /* write cycle time  tGPR */
{ return aC_4() + dGPRwrite() + delta;
}

/* ---------------------------------------------------------- */
/*   delay of the forwarding engine                           */
/* ---------------------------------------------------------- */
int dforw_Do()                                /* -> Dout */
{ return dxor + 2*dand + 3*dmux + dzero(ld(var.A));
}

int dforw_top()                               /* -> top */
{ return dxor + 4*dand + 3*dmux + dzero(ld(var.A));
}

int dforw_Dat()                         /* data -> Dout */
{ return 3*dmux;
}

int aABin()
{ int aDataIn;

  if (DLXp == 0)
   return dGPRread();
  else
  { aDataIn = max(max(aEX(), aC_4()), dGPRread()) +dforw_Dat();
    return max(aDataIn, dforw_Do());
  }
}

/* ---------------------------------------------------------- */
/*     ID stage: IRenv, CAddr, PCenv, GPRread                 */
/* ---------------------------------------------------------- */
/*---- operand fetch ---------------------------------------- */
int tABread()                              /* read cycle time */
{ return aABin() + dff +delta;
}

/*---- IR Environment (Mealy) ------------------------------- */
int aIRco()
{ return acsMealy() + dmux;
}

int tIRenv()
{ return aIRco() + dff + delta;
}

/*---- Destination Address CAddr ---------------------------- */
int tCAddr()
{ return acsMealy() + 2*dmux + dff + delta;
}

/* --- PC environment (delayed PC) -------------------------- */
int adpc()                   
{ return dmux;
}

int aPCglue()                   
{ int delay;

  delay = dor + dand + dxor;          /* dPCglue */
  return max(acsMealy(), aABin() +dfzero(var.n)) + delay;
}

int tPCenv()                   
{ int delay;

  delay = max(dincf(var.n -2), aIRco() + dCLAf(var.n));
  delay = max(max(delay, aABin()), aPCglue());
  return delay + 3*dmux + dff + delta;
}

int tID()                   
{ return max(max(tPCenv(), tIRenv()), max(tCAddr(), tABread()));
}

/*== cycle time: data paths DP ================================*/
int tDP()
{ return max( max(tWB(), tEX()), tID());
}
 
/*=============================================================*/
/*   Delay / Cycle Time of the memory stages: IF, M            */
/*=============================================================*/
int aDMC()                  /* control signals of DMC */
{ int delay;

  delay = acsMoore();                      /* inputs */

  return  delay + max(ddecf(2), dinv + dand) + 2*(dand+dor);
}

int aIMC()                  /* control signals of IMC */
{ return adpc();
}

int tIF()
{ return aIMC() + var.dImem + dff + delta;
}

int aMbusy()
{ return max(aIMC() +var.dIstat, aDMC() + var.dstat);
}

int tDM()
{ return aDMC() + var.dmem + dff + delta;
}

/* --- Memory cycle time tMM -------------------------------- */
int tMM()
{ return max(tIF(), tDM());
}

/*============================================================*/
/*   Control Environment (including pAuto, Cce, IMC, DMC)     */
/*============================================================*/
int tpauto()
{ int delay;

  stage = 1; con = conMealy; 
  delay = tauto();                      /* Mealy(ID) */

  con = conMoore; stage = 2;            /* pMoore(EX-WB) */
  return max(delay, tauto()); 
}

int tCONbuf()
{ return (DLXp ==0) ? tpauto() : tpauto() + dand;
}

int tstall()
{ int aCE, aDhaz;

  if (DLXp <= 1)
     aCE = aMbusy() + dnor + dor;
  else
  { aDhaz = dforw_top() +dand +2*dor; 
    aCE = max(aMbusy(), aDhaz) + max(dnor, dinv) +dand +dor; 
  }

  return aCE +3*dand +max(dff, dGPRwrite()) +delta; 
}

/* ----- Control Unit --------------------------------------- */
int tCON()
{ return max(tCONbuf(), tstall());
}

/* ---------------------------------------------------------- */
/*     Cost and cycle time of the DLX Design                  */
/* ---------------------------------------------------------- */
int cDLX()
{ return cDP() + cCON();
} 

int tDLX()
{ return max(tDP(), tCON());
} 

int cCONrel()
{ return (int)((100.0* cCON())/ cDLX());
}

int cGPRrel1()
{ return (int)((100.0* cGPR())/ cDP());
}

int cGPRrel2()
{ return (int)((100.0* cGPR())/ cDLX());
}


/* === Print functions ====================================== */
void print_line()
{
  printf(
  "---------------------------------------------------------\n");
  return;
}

void print_header(str1,str2)
char *str1, *str2;
{
  printf("%-14s \t      IDmem / IDstat \n",str1);
  printf("%-14s \t    0/0     25/5    0/20    50/0\n",str2);
  print_line();
  return;
}

void print_value(str,func)
char *str;
int (*func)();
{ printf("%14s:",str); 
  var.dstat = var.dIstat = 0;  var.dmem = var.dImem = 0;  
  printf("\t%7d",(*func)());
  var.dstat = var.dIstat = 5;  var.dmem = var.dImem = 25;  
  printf("\t%7d",(*func)());
  var.dstat = var.dIstat =20;  var.dmem = var.dImem = 0;  
  printf("\t%7d",(*func)());
  var.dstat = var.dIstat = 0;  var.dmem = var.dImem = 50;  
  printf("\t%7d\n",(*func)());
  return;
}


/* ---------------------------------------------------------- */
/*     Evaluation of the design                               */
/* ---------------------------------------------------------- */

void main()
{
 char my_str[200];             /* temporary strings */
 int i, x, y;                  /* loop variables */
 int c1, c2, t1, t2;           /* cost, time values */
 double res, q;                /* result, quality parameter */

 for (DLXp =0; DLXp<=2; DLXp++)
 {
  init_techno(MOTOROLA);        /* select technology */

  var.n      = 32;              /* width of the data paths */
  var.A      = 32;              /* number of GP-Registers */
  var.dmem  = var.dImem = 25;  /* memory access times */
  var.dstat = var.dIstat = 5;  /* memory status times */

  printf("\n\n\n Cost Delay Analysis of the DLX Design \n");
  printf(" Pipelining, hardwired control, no interrupt\n");
  printf(" Technology:  %s, ", name_techno());

  strcpy(my_str,"CLA"); 
  printf(" %s Adder, Design DLXp = %d \n", my_str, DLXp);
  printf(
   " width n = %d,  Reg-number = %d,  Mem status time = %d\n",
   var.n,var.A,var.dstat);

/* ---------------------------------------------------------- */
/*  Cost of the DP environments and the control               */
/* ---------------------------------------------------------- */

  print_line();              /* cost analysis */
  print_header("cost","");
  print_value("EXenv",cEXenv);
  print_value("SH4Le",cSH4Le);
  print_value("IRenv",cIRenv);
  print_value("CAddr",cCAddr);
  print_value("GPRenv",cGPR);
  print_value("PCenv",cPCenv);
  print_value("Buffer",cBuffer);
  print_line();
  
  print_value("stall",cstall);
  print_value("MDC",cDMC);
  print_value("pauto",cpauto);
  print_value("CONbuf",cCONbuffer);
  print_line();

  print_value("DP",cDP);
  print_value("Control",cCON);
  print_value("DLX",cDLX);
  print_line();

  print_value("CON/DLX [%]",cCONrel);
  print_value("GPR/DP [%]",cGPRrel1);
  print_value("GPR/DLX [%]",cGPRrel2);
  print_line();
  
/* ---------------------------------------------------------- */
/*    Cycle time and delay of DP environments                 */
/* ---------------------------------------------------------- */
  print_line();              /* cycle time analysis */
  print_header("delay","  ");

  print_value("tID",tID);
  print_value("tPCenv",tPCenv);
  print_value("tABread",tABread);
  print_value("tEX",tEX);
  print_value("tWB",tWB); 
  print_value("tDP",tDP);
  print_line();
  print_value("acsMealy",acsMealy);
  print_value("aABin",aABin);
  print_value("aPCglue",aPCglue);
  print_line();
  print_value("aMbusy",aMbusy);
  print_value("tCONbuf",tCONbuf);
  print_value("tstall",tstall);
  print_value("tCON",tCON);
  print_line();
  print_value("tDLX",tDLX);
  print_line();
  print_value("tIF,M",tMM);
  print_line();
  printf("\n\n\n");
 }
}
