/* =========================================================== */
/* pdlxcF.c    pipelined DLX core with FPU                     */
/*             with interrupt handling, with caches, CLA adder */
/* ----------------------------------------------------------- */
/*  last update: 99/08/13 SMM                                  */
/* =========================================================== */
/*  BUS =  0: mbusy                                            */
/*         1: caches   (line, sector, bytes)                   */
/* ----------------------------------------------------------- */

#include <stdio.h>
#include <math.h>
#include "techno.h"
#include "macro.h"
#include "EXmacro.h"
#include "hardctr.h"
#include "cache.h"
#include "evaluate.h"

struct variable_parameters var;

struct FPU_parameters {
  int cost;             /* cost of the FPU */
  int output;           /* accumulated delay of FPU outputs */
  int cycle;            /* cycle time of the FPU */
};

struct FPU_parameters FPU = {104022, 50, 98};


int stage;
/* ----------------------------------------------------------- */
/* identifies the two control automata                         */
/* stage = 1: Mealy       2: Moore                             */
/* ----------------------------------------------------------- */
/* parameter sets                                              */
/* states, zeta - in, out - monom, lmax, lsum - numax, nusum - */
/* fanmax, fansum - type, binary - aIN, aclr                   */
/* ----------------------------------------------------------- */
struct control_parameters conMealy =
       {1,0, 15,13, 21,15,147, 5,21, 1,0, 2,0, 0,0};
struct control_parameters conMoore = 
       {44,6, 15,48, 53,15,374, 30,342, 3,53, 1,0, 0,0};

/* Memory Controller: with sectored cache */
/* cost and cycle time */
struct control_parameters conIC1 =
    {8,3,6,11, 10,3,17, 5,29, 3,16, 1,0,0,0};
struct control_parameters conDC1 =
    {11,4,8,18, 18,3,35, 5,42, 3,20, 1,0,0,0};
/* delay of the Mealy signals */
struct control_parameters conID2 =
    {11,4,8,18, 18,2,35, 3,42, 3,20, 3,0,0,0};

struct cache_parameters L1c;
struct cache_parameters L1cache = { 22, 5, 2, 3, 2, 0};
/*       { 22,  5, 2, 3, 2, 0};       32B  2KB/2 */
/*       { 19, 10, 0, 3, 1, 0};        8B  8KB */
/*       { 18, 10, 1, 3, 1, 0};       16B 16KB */
/*{tag, line, sector, boff, way, wr-through} */



/*------------------------------------------------------------*/
/*   Result shift register rsr(f,n)                           */
/*------------------------------------------------------------*/
int cRSR(f,n)
int f,n;
{ return f*n*(cff +cand +cmux) +f*cor +cinv;
}
 
int dRSR()
{ return dand + dmux;
}

/*=============================================================*/
/*   Cost of the data paths and its environments               */
/*=============================================================*/
/*---- IR Environment ----------------------------------------*/
int cIRenv()
{ return 15*cmux +var.n*cff;
}

/*---- CAddr (destination address Cad) -----------------------*/
int cCAddr()
{ return 2* ld(var.A)* cmux;
}

/*---- Daddr (register file addresses) -----------------------*/
int cDaddr()
{ return cCAddr() + 4*5*cmux;
}
 
/*---- GPR register file -------------------------------------*/
int cGPR()
{ return  cram2(var.A, var.n)
          + 2* (czero(ld(var.A)) + cinv + var.n* cand);
}
 
/*---- Special register file SF ------------------------------*/
int cSF(K,n)
int K,n;
{ return  K*n* (cff + cmux) + 2*cdecf(ld(K)) + cand *K *(n+1)
          + (K + n* ctree(K)) *cor;
}
 
/*---- register file SPR -------------------------------------*/
int cSPR()
{ return  cSF(var.S, var.n)
          +3 * var.n * cmux              /* SPRsel */
          +5 * cor                       /* FCRsel */
          +czero(ld(var.S)) +2*cand +cinv;  /* SPRglue */
}
 
/*---- register file FPR -------------------------------------*/
int cFPR()
{ return  2*cram2(var.A/2, var.n)
          +2*cand + 2*cor + cinv;        /* FPRcon */
}
 
/*---- register file environment RFenv -----------------------*/
int cRFenv()
{ return  cFPR() + cSPR() + cGPR();
}
 
/*---- execute environment EXenv -----------------------------*/
int cFPXtr()
{ return var.n* 6*cmux;
}
 
int cEXenv()
{ return var.n*(2*cmux+6*cdriv) +cALUenv()+cshenv()    /* FXU */
         + FPU.cost + cFPXtr()
	 + (2*var.n + 4*69)* cdriv;
}


/*------------------------------------------------------------*/
/*   Cause environment                                        */
/*------------------------------------------------------------*/
/*---- Cause collection --------------------------------------*/
int cCAcol()
{ return  6*cand +cor +13*cff + cRSR(5,3);
}

/*---- Cause processing --------------------------------------*/
int cCApro()
{ return  25*cand +cor*ctree(var.n) +34*cff
	  +3*cor +cand +cinv;                  /* CAtype */
}

/*---- Cause environment--------------------------------------*/
int cCAenv()
{ return  cCApro() + cCAcol();
}

/*-------------------------------------------------------------*/
/*   PC environment (delayed PC)                               */
/*-------------------------------------------------------------*/
/*---- PC glue logic -----------------------------------------*/
int cPCglue()
{ return 3*cor +2*cand +cxnor +cxor +czero(var.n);
}

/*---- PC data paths -----------------------------------------*/
int cPCenv()
{ return var.n* (4*cff + 6*cmux) + cCLAf(var.n) 
	  + cincf(var.n-2) + cPCglue();
}

/*---- FPemb  embedding of floating point data ---------------*/
int cFPemb()
{ return 2*( 2*cor + cinv + 2*var.n*cmux);
}

/*---- GPR forwarding data paths (+control, 2x Forw3, +top) --*/
int cGforw()
{ int cost, ldA;

  ldA = ld(var.A);
  cost = cEQ(ldA) + 3*cand + var.n*cmux;

  cost = 3*cost + 3*cand + 2*cinv + ctree(ldA)*cor;
  return 2*cost;
}

/*-------------------------------------------------------------*/
/*   SPR forwarding: general x-stage,  + SR, EPC and EDPC      */
/*-------------------------------------------------------------*/
/*---general x-stage SPR forwarding ---------------------------*/
int csfor(x)
int x;
{ return var.n*x*cmux + x*(cEQ(ld(var.S)) +2*cand);
}

/*--- Forwarding of SR and EPC --------------------------------*/
int cSforw()
{ return csfor(1) +3*csfor(3);
}

/*-------------------------------------------------------------*/
/*   FPR forwarding: circuit FFor forwarding of 2 double words */
/*-------------------------------------------------------------*/
/*--- selection of one double word ----------------------------*/
int cFforSel()
{ return 2* (3*var.n*cmux + cor);
}
 
/*--- generating of the hit signals ---------------------------*/
int cFforHit()
{ return cinv +2*cor +7*(cEQ(ld(var.A/2)) +6*cand +2*cor +cinv);
}
 
/*--- circuit FFor which forwards operands Fa and FB ----------*/
int cFFor()
{ int ffor;

  ffor = cFforHit() +2*(7*cand +6*cinv +cand*cPP(6))
	 +6*(cor + cinv + cand) + ctree(6)*cor;
  return 2* (ffor + cFforSel()); 
}
 
/*--- Cost of the Forwarding engine FORW --------------------*/
int cFORW()
{ return cFFor() + cSforw() + cGforw();
}
 
/*------------------------------------------------------------*/
/*   Buffer environment                                       */
/*------------------------------------------------------------*/
int cBuffer()
{ return  cRSR(5,123) + 123*cff + 27*cff;
}

/*-------------------------------------------------------------*/
/*   Memory environments IMenv / DMenv                         */
/*-------------------------------------------------------------*/
int cIMenv()                        /* IMenv */
{
  if (var.BUS==0 )
    return 0;  /* no cache */
  else
    return  2*cmux*var.n + CchIF();
}

int cDMenv()                        /* IMenv */
{
  if (var.BUS==0 )
    return 0;  /* no cache */
  else
    return  cmux*var.n + CchIF();    /* 64-bit output */
}

int cMif()                        /* Mif */
{ return (var.BUS==0 ? 0 : var.n*(2*cdriv + cmux));
}

int cMenv()                        /* IMenv */
{ return cIMenv() + cDMenv() + cMif();
}


/*-------------------------------------------------------------*/
/*   whole DLX data paths                                      */
/*-------------------------------------------------------------*/
int cDP()
{ return cPCenv() + cIRenv() + cEXenv() + cRFenv() + cSH4Lif()
	 + cCAenv() + cDaddr() + cMenv() + cFPemb() + cBuffer() 
	 + cFORW() + (5 + 10)*var.n*cff + 2*5*cff;
}


/*=============================================================*/
/*   Cost of the control: stall, automata, IMC, DMC            */
/*=============================================================*/
int cIMC()
{ return cor;
}

int cDMC()
{ return cdecf(3) + 32*cand + 20*cor + 6*cinv +4*cnor +8*cff;
}

/* --- Memory Interface Control MifC ------------------------- */
int cMifC()
{ int cost;

  if (var.BUS==0)
    return 0;
  else
  { con = conIC1;  cost = cauto();
    con = conDC1;  
    return  cost + cauto();
  }
}

/* --- Computation of the hazard flags ----------------------- */
int CdhazIEEEf()
{ return cEQ(4) + cor*ctree(5) + 11*cand + 2*cor + cnor + cinv;
}

int CdhazEDPC()
{ return 3*cEQ(ld(var.S)) +7*cand +2*cor;
}

int CdhazS()
{ return CdhazIEEEf() + CdhazEDPC() + cor;
}

int CdhazFP()
{ return 2*cand + cor;
}

int CdhazFX()
{ int cost;
 
  cost = 2*cinv + 4*cand + 2*cor;     /* old: dhazA, dhazB */
  return  cost +2*cand + cor; 
}

int cCE()
{ int dhaz, RSRhaz, busy;

  dhaz = CdhazS() +CdhazFP() +CdhazFX() +cor +cnor;
  RSRhaz = 2*cand +4*cor;
  busy = 2*cand +3*cnor;
  return dhaz +RSRhaz +busy +4*cor +cnor +cinv +cnand +3*cand;
}

int cstall()
{ int core;

  core = 25*cand +cor +2*cff +cRSR(21,1);
  return core +14*cand +4*cor +cinv +cnand +cnor;
}

/* --- control automata -------------------------------------- */
int cpauto()
{ int cost;

  con = conMealy; stage = 1; cost = cauto();   /* Mealy */
  con = conMoore; stage = 2; cost += cauto();  /* pMoore(EX-WB) */
  return cost;
}

/* ------------------------------------------------------------ */
/* precomputed control                                          */
/* valid flags, buffers for the Moore signals, DIV control      */
/* ------------------------------------------------------------ */
int CpreCON()
{ int conRSR, DivCON;

  DivCON = cff*7 +cmux*6 +cand +ctree(16)*cor + 19*cor;
  conRSR = cinv +5*cor +90*(cand +cmux +cff);
  return DivCON + conRSR +cand +cinv +cff*(8+6);
}


int cCON()
{ return cIMC() +cDMC() +cMifC() +cstall() +cCE() 
         +cpauto() +CpreCON();
}


/*=============================================================*/
/*=== Delay and Cycle Time ====================================*/
/*=============================================================*/
/*=============================================================*/
/*   Control Automata: Delay of the control signals            */
/*=============================================================*/
int acsMealy()                 /* Mealy signals: ID stage */
{ con = conMealy;
  return aOut();
}

int acsMoore()                  /* Moore signals */
{ con = conMoore;
  return aOut();
}

int aJISR()                     /* JISR = jisr.4 ^ full.4 */
{ return max(dand, dnand);
}

/*=============================================================*/
/*   Delay / Cycle Time of the data paths environments         */
/*=============================================================*/
/* --- execute stage EXenv ----------------------------------- */
int aFXU()
{ return acsMoore() +dmux +max(dALUenv(), dALUe2()+dshenv())
	      + ddriv;
}

int aFPXtr()
{ return acsMoore() +dALUe2()+dshenv() + 2*ddriv;
}

int aEX()
{ return max( aFXU(), max(FPU.output, aFPXtr()) + ddriv);
}

int aALUovf()
{ return acsMoore() +dmux +dALUovf();
}

int tEX()
{ return max(FPU.cycle, aEX() + dff + delta);
}

/* ---------------------------------------------------------- */
/*  WB stage: SH4L GPR (# Registers: var.A, width: var.n)     */
/*                 SPR (# Registers: var.S, width: var.n)     */
/* ---------------------------------------------------------- */
/*---- Special register file SF ------------------------------*/
int dSFdout(K)                /* std. data output */
int K;
{ return  ddecf(ld(K)) + dand +dtree(K) *dor;
}

int dSFw(K)                   /* delay on write */
int K;
{ return  ddecf(ld(K)) + dand +dor +dff;
}

int dSFw_ce()                 /* SFw[],SFw/r -> ce*/
{ return  dand +dor;
}

/*---- register file SPR -------------------------------------*/
int aSPRsin()                   /* delay of std output Sin */
{ return dSFdout(var.S);
}

int tSPR()                   /* write cycle time  of SPR */
{ return max(dzero(ld(var.S)), aJISR()) 
	   +2*dand +dmux +dSFw(var.S) +delta;
}

/*---- register file GPR -------------------------------------*/
int dGPRread()                   
{ return max(dram2(var.A, var.n), dzero(ld(var.A)) +dinv) 
	 + dand;
}

int dGPRwrite()                   
{ return dram2(var.A, var.n);
}

/*---- register file FPR -------------------------------------*/
int dFPRread()                   
{ return dram2(var.A/2, var.n);
}

int dFPRwrite()                   
{ return dram2(var.A/2, var.n);
}

int dFPRcon()
{ return dinv +dor +dand;
}

int aRFin()                  /* input data of GPR and FPR */
{ return dSH4Lie();
}

int tWB()                   /* write cycle tGPR, tSPR, tFPR */
{ int aFPR, delay;

  aFPR  = max(aRFin(), dFPRcon()) + dFPRwrite();
  delay = max(aFPR, aRFin()+ dGPRwrite()) + delta;
  return max(tSPR(), delay);
}

/* ---------------------------------------------------------- */
/*   delay of the forwarding engine GPR                       */
/* ---------------------------------------------------------- */
int dforw_Do()                                /* -> Dout */
{ return dxor + 2*dand + 3*dmux + dzero(ld(var.A));
}

int dforw_top()                               /* -> top */
{ return dxor + 4*dand + 3*dmux + dzero(ld(var.A));
}

int dforw_Dat()                         /* data -> Dout */
{ return 3*dmux;
}

int aABin()
{ int aDataIn;

  aDataIn = max(max(aEX(), aRFin()), dGPRread()) +dforw_Dat();
  return max(aDataIn, dforw_Do());
}

/* ---------------------------------------------------------- */
/*   delay of the forwarding engine for FPR                   */
/* ---------------------------------------------------------- */
/* ---- select double word: Data -> dout -------------------- */
int dFforSel()
{ return 3*dmux;
}

/* ---- delay of the hit signals ---------------------------- */
int dFforHit()
{ return 2*dand +max(dEQ(ld(var.A/2)), dinv+dor);
}

int dFfor_Do()                                /* -> Dout */
{ return dFforSel() +dFforHit();
}

int dFfor_dhaz()                              /* -> dhaz */
{ return dFforHit() +dinv +dand*(dPP(6)+2) +dor*(1+dtree(6));
}

int aFABin()
{ int aDataIn;

  aDataIn = max(max(aEX(), aRFin()), dFPRread());
  return max(aDataIn +dFforSel(), dforw_Do());
}


/* ---------------------------------------------------------- */
/*   delay of the forwarding engine SPR                       */
/* ---------------------------------------------------------- */
int dSfor_hit()                               /* -> hit */
{ return dEQ(ld(var.S)) + dand;
}

int dSfor_Do(x)                               /* -> Dout */
int x;
{ return dSfor_hit() + x*dmux;
}

int dSfor_Dat(x)                         /* data -> Dout */
int x;
{ return (x)*dmux;
}

int aSRfor()
{ return max(aJISR() + dmux +dSfor_Dat(1), dSfor_Do(1));
}

int aEPCfor()
{ return max(dSfor_Do(3), 
             max(aFXU(), aJISR()+dmux) +dSfor_Dat(3));
}

int aFCCfor()
{ return aEPCfor();
}

int aSin()                     /* reading operand SPR(Sas) */
{ return max(acsMealy()+dSfor_Do(3), 
             max(aFXU(), aJISR()+dmux) +dSfor_Dat(3));
}

/*---- FPemb  embedding of floating point data ---------------*/
int dFPemb()
{ return dmux;
}

/* ---------------------------------------------------------- */
/*     ID stage: IRenv, CAddr, PCenv, GPRread                 */
/* ---------------------------------------------------------- */
/*---- operand fetch ---------------------------------------- */
int tABread()                        /* read operands A, B */
{ return aABin() + dff +delta;
}

int tSread()                         /* read operand S */
{ return aSin() + dff +delta;
}

int tFABread()                       /* read operands FA FB */
{ return aFABin() + dFPemb() +dff +delta;
}

int tOPread()                        /* read all operands */
{ return max(tABread(), max(tSread(), tFABread() ));
}

/*---- IR Environment (Mealy) ------------------------------- */
int aIRco()
{ return acsMealy() + dmux;
}

int tIRenv()
{ return aIRco() + dff + delta;
}

/*---- Destination Address CAddr ---------------------------- */
int aDAddr()
{ return acsMealy() + 2*dmux;
}

int tDAddr()
{ return aDAddr() + dff + delta;
}

/* --- PC environment (delayed PC) -------------------------- */
int adpc()                   
{ return aJISR() + 2*dmux;      /* EDPC never forwarded */
}

int aPCglue()                   
{ int delay, delay2;

  delay = dor + dand + dxor;          /* dPCglue */
  delay2 = max(aFCCfor(), aABin() +dfzero(var.n));
  return max(acsMealy(), delay2) + delay;
}

int tPCenv()                   
{ int delay;

  delay = max(dincf(var.n -2), aIRco() + dCLAf(var.n));
  delay = max(max(delay, aABin()), max(aPCglue(), aEPCfor()) );
  return delay + 3*dmux + dff + delta;
}

int tID()                   
{ return max(max(tPCenv(), tIRenv()), max(tDAddr(), tOPread() ));
}

/*=============================================================*/
/*   Delay / Cycle Time of the memory stages: IF, M            */
/*=============================================================*/
/* --- Set Delay of cacheIF inputs --------------------------- */
int INcache()
{
  cacheIF.A = adpc() + dmux;
  cacheIF.Din = cacheIF.csig =0;
}

int aDMC()                  /* control signals of DMC */
{ int aDMC_MW, delay;

  aDMC_MW = aJISR() + 2*dand + dnor + dor;
  delay = acsMoore() + max(ddecf(3), 2*dand +dor) +2*(dand+dor);

  return max(delay, aDMC_MW);
}

int aDMC_BE()
{
  if (var.BUS == 0)
    return aDMC();
  else 
    return 0;             /* with cache: MBW are buffered */
}

/* --- Memory Interface Control MifC ------------------------- */
int aMifC()
{ return (var.BUS==0 ? 0 : dmux);     /* Bus handshake signals */
}

int aIMC()
{ return adpc() + dor;                /* imal */
}
 
/* ---------------------------------------------------------- */
/*    Memory cycle times without caches                       */
/* ---------------------------------------------------------- */
int tIF()
{ return adpc() + var.dImem + dff + delta;
}

int tDM()
{ return aDMC() + var.dmem + dff + delta;
}

/* --- memory cycle time "TMaccess"-------------------------- */
int tMM()
{ return max(tIF(), tDM());
}

/* ---------------------------------------------------------- */
/*    Memory cycle times with cache                           */
/* ---------------------------------------------------------- */
int tM()                                 /* t_Maccess */
{ int delay;

  if (var.BUS==0)
   return tMM();
  else
  { INcache();  
    delay = DchIFM_ch() + ddriv;
    delay = max(max(delay, aMifC()), aDMC_BE());
    return delay +var.dbus +dDRAM(var.Mword,8) +delta;
  }
}

int tcacheIF()                               /* t_cacheIF */
{ int delay;

  INcache(); delay = TchIF();
  return (var.BUS==1 ? delay : 0); 
}
 
int tMrbst()                               /* t_Mrburst */
{ int delay;
 
  INcache();
  delay = max(DchIFM_ch(), DchIFM_Dout()+ dff)
                + ddriv +var.dbus +delta;

  return (var.BUS==1 ? delay : 0); 
}

int tMreq()                               /* t_Mrequest */
{ int delay;
 
  INcache();
  delay = AchIF_MAdr();
  delay = delay + ddriv +var.dbus + var.dMhsh + dff +delta;
  return (var.BUS==1 ? delay : 0); 
}
 
int tcachRD()                               /* t_$rd */
{ int delay;
 
  INcache();
  delay = AchIF_Dout()+ dff;
  return (var.BUS==1 ? delay : 0); 
}

 
/* --- Acc. delay of dbusy/ibusy----------------------------- */
int aMbusy()                       /* signals ibusy, dbusy */
{int delay;
 
  switch(var.BUS)
  { case 0:                      /* no bus interace -- mbusy */
      return max(aIMC() +var.dIstat, aDMC() + var.dstat);
 
    default:                      /* aMifC(mbusy)>0 */
      INcache();
      delay = max(AchIF_hit(), max(aDMC(), aIMC()));
      conID2.aIN = max(delay, var.dbus);   con = conID2;
      return aOut();
  }
}
 
 
/*=============================================================*/
/*   Cause environment: collection and processing              */
/*=============================================================*/
int aCAcol()                             /* A_CAcol(CA.4) */
{ if (var.BUS==0)
    return aDMC() +max(dor, var.dstat);
  else
    return aDMC() +dor;
}

int tCAcol()                         /* CAcol: register update */
{ int dfpu, delay;

   dfpu  = max(acsMealy(), FPU.output + ddriv);
   delay = max(dfpu, aALUovf() + dand);

if (var.BUS==0)
   return max(adpc()+var.dIstat, delay) + dff + delta;
  else
   return max(aIMC(), delay) + dff + delta;
}

int tCApro()                         /* CApro: register update */
{ return max(aSRfor(), aCAcol()) 
	   +dand + dor*dtree(var.n) + dff + delta;
}

int tCAenv() 
{ return max(tCAcol(), tCApro());
}

/*== cycle time: data paths DP ================================*/
int tDP()
{ return max( max(tWB(), tEX()), max(tID(), tCAenv()) );
}
 
/*============================================================*/
/*   Control Environment (including pAuto, Cce, IMC, DMC)     */
/*============================================================*/
int tpauto()
{ int delay;

  stage = 1; con = conMealy; 
  delay = tauto();                      /* Mealy(ID) */

  con = conMoore; stage = 2;            /* pMoore(EX-WB) */
  return max(delay, tauto()); 
}

int tpreCON()
{ int tDIVcon;

  tDIVcon = dand +dor*dtree(16) +dmux +dff +delta;
  return max(tpauto() + dand, tDIVcon);
}

int dIEEEfhaz()
{ int delay;

  delay = max(dEQ(ld(var.S)), (1+dtree(5))*dor);
  return 2*dand + max(delay, dSfor_hit() +2*dor +dnor);
}

int dEDPChaz()
{ return dEQ(ld(var.S)) +2*dand +2*dor;
}

int AdhazS()
{ return aDAddr() +dor +max(dEDPChaz(), dIEEEfhaz());
}

int AdhazFP()
{ return acsMealy() +dFfor_dhaz() +dand +dor;
}

int AdhazFX()
{ return dforw_top() +2*dand +2*dor;
}

int aRSRhaz()
{ return dand +3*dor;
}

int aNOTdhaz()
{ return max(AdhazS(), max(AdhazFP(), AdhazFX())) +dor +dnor;
}

int tstall()
{ int Anotbusy, aCE, Aue, delay;

  Anotbusy = (var.BUS==0 ? 2*dnor +dand : dnor);
  Anotbusy = Anotbusy +aMbusy();

  aCE = max(Anotbusy, max(aNOTdhaz(), aRSRhaz())) +dinv +2*dand +dor; 
  Aue = aCE +dRSR() +dand;
 
  delay = max(dGPRwrite(), dFPRcon() +dFPRwrite());
  return Aue +2*dand + max(dSFw_ce() +dff, delay) +delta; 
}

/* --- Memory Interface Control MifC ------------------------- */
int tMifC()          /* automaton in MifC */
{ int delay;

  switch(var.BUS)
  { case 0:                      /* no bus interace -- mbusy */
      return 0;

    default:                     /* sectored cache */
      INcache();
      delay = max(aIMC(), aDMC());
      delay = max(delay, AchIF_hit());
      conDC1.aIN = max(delay, var.dbus);
      conDC1.aclr = 0;   con = conDC1;
      delay = tauto();           /* update of MifC automaton */
                                 /* buffering of MBW */
      return  max(delay, aDMC()+dff+delta);
  }
}


/* ----- Control Unit --------------------------------------- */
int tCON()
{ return max(max(tpreCON(), tstall()), tMifC());
}

/* ---------------------------------------------------------- */
/*     Cost and cycle time of the DLX Design                  */
/* ---------------------------------------------------------- */
int cDLX()
{ return cDP() + cCON();
} 

int tDLX()
{ return max(tDP(), tCON());
} 

int cCONrel()
{ return (int)((100.0* cCON())/ cDLX());
}


/* === Print functions ====================================== */
void print_line()
{
  printf(
  "---------------------------------------------------------\n");
  return;
}

void print_header(str1,str2)
char *str1, *str2;
{
  printf("%-14s \t      Dmem,bus / Dstat,Mhsh \n",str1);
  printf("%-14s \t    0/0     15/10   0/20    50/0\n",str2);
  print_line();
  return;
}

void print_value(str,func)
char *str;
int (*func)();
{ printf("%14s:",str);
  var.dstat = var.dIstat = var.dMhsh = 0;
  var.dmem = var.dImem = var.dbus = 0;

  printf("\t%7d",(*func)());
  var.dstat = var.dIstat = var.dMhsh = 10;
  var.dmem = var.dImem = var.dbus = 15;
 
  printf("\t%7d",(*func)());
  var.dstat = var.dIstat =var.dMhsh = 20;
  var.dmem = var.dImem = var.dbus = 0;
 
  printf("\t%7d",(*func)());
  var.dstat = var.dIstat = var.dMhsh = 0;
  var.dmem = var.dImem = var.dbus = 50;
  printf("\t%7d\n",(*func)());
  return;
}

void cacheheader(str1,value)
char *str1;
int value;
{
  printf("%-11s \t      Line size [Bytes] \n",str1);
  printf("%-11d \t    8       16     32       64    128 \n",value);
  forall_line();
  return;
}
 
void fL1_print(str,func)
char *str;
int (*func)();
{ printf("%11s:",str);
  cache = L1c;
  printf("\t%7d",(*func)());
 
  cache.sector +=1; cache.line -=1;     /* 16B*/
  printf("\t%7d",(*func)());
 
  cache.sector +=1; cache.line -=1;     /* 32B*/
  printf("\t%7d",(*func)());
 
  cache.sector +=1; cache.line -=1;     /* 64B*/
  printf("\t%7d",(*func)());
 
  cache.sector +=1; cache.line -=1;     /* 128B*/
  printf("\t%7d\n",(*func)());
  return;
}


/* ---------------------------------------------------------- */
/*     Evaluation of the design                               */
/* ---------------------------------------------------------- */

void main()
{
 int i, way, size;
 char my_str[200];             /* temporary strings */

 init_techno(MOTOROLA);        /* select technology */

 cache = L1cache;

 var.n      = 32;              /* width of the data paths */
 var.A      = 32;              /* number of GP-Registers */
 var.S      = 6;               /* number of SP-Registers */
 var.dmem  = var.dImem = 25;   /* memory access times */
 var.dstat = var.dIstat = 5;   /* memory status times */
 var.dbus   = 15;              /* delay of the memory bus */
 var.dMhsh  = 10;              /* memory delay: reqp, Brdy*/
 var.Mword  = 16 *1024 * 1024; /* memory size in words */

 for(i=0; i<2; i++)
 {
   var.BUS = i;
   cache = L1cache;

   printf("\n DLX Design: pipelining, with interrupts \n");
   printf(" Memory interface: %d; delay dBUS = %d \n",
          var.BUS,var.dbus);
   printf(
" Cache: lines= %d, sector= %d, offset= %d, way= %d, Write: %d \n",
    cache.line,cache.sector,cache.boff,cache.way,cache.wrpol);

/* ---------------------------------------------------------- */
/*  Cost of the DP environments and the control               */
/* ---------------------------------------------------------- */

   print_line();              /* cost analysis */
   print_header("cost","");
   print_value("EXenv",cEXenv);
   print_value("SH4Le",cSH4Lif);
   print_value("IRenv",cIRenv);
   print_value("Daddr",cDaddr);
   print_value("RFenv",cRFenv);
   print_value("CAenv",cCAenv);
   print_value("PCenv",cPCenv);
   print_value("FORW",cFORW);
   print_value("Buffer",cBuffer);
   print_value("Menv",cMenv);
   print_line();
   print_value("stall",cstall);
   print_value("MDC",cDMC);
   print_value("IDC",cIMC);
   print_value("MifC",cMifC);
   print_value("pauto",cpauto);
   print_value("preCON",CpreCON);
   print_value("cCE",cCE);
   print_line();
   print_value("DP",cDP);
   print_value("Control",cCON);
   print_value("DLX",cDLX);
   print_line();
   print_value("CON/DLX [%]",cCONrel);
   print_line();
  
/* ---------------------------------------------------------- */
/*    Cycle time and delay of DP environments                 */
/* ---------------------------------------------------------- */
   print_line();              /* cycle time analysis */
   print_header("delay","  ");

   print_value("tID",tID);
   print_value("tPCenv",tPCenv);
   print_value("tOPread",tOPread);
   print_value("tEX",tEX);
   print_value("tWB",tWB); 
   print_value("tDP",tDP);
   print_line();
   print_value("acsMealy",acsMealy);
   print_value("aABin",aABin);
   print_value("aFABin",aFABin);
   print_value("aSin",aSin);
   print_value("aPCglue",aPCglue);
   print_line();
   INcache();
   print_value("Ahit",AchIF_hit);
   print_value("aMbusy",aMbusy);
   print_value("AdhazFX",AdhazFX);
   print_value("AdhazFP",AdhazFP);
   print_value("AdhazS",AdhazS);
   print_value("aNdhaz",aNOTdhaz);
   print_value("aRSRhaz",aRSRhaz);
   print_value("tpreCON",tpreCON);
   print_value("tstall",tstall);
   print_value("tMifC",tMifC);
   print_value("tCON",tCON);
   print_value("tDLX",tDLX);
   print_line();
   print_value("T_Mreq",tMreq);
   print_value("T_Mrbst",tMrbst);
   print_value("T_cache",tcacheIF);
   print_value("T_cachRD",tcachRD);
   DRAMfac = 4;  print_value("T_M(4)",tM);
   DRAMfac = 8;  print_value("T_M(8)",tM);
   DRAMfac = 16; print_value("T_M(16)",tM);
   print_line();
  }

/* ---------------------------------------------------------- */
/*    Variable cache sizes                                    */
/* ---------------------------------------------------------- */
  var.BUS = 1;
  var.dbus   = 15;              /* delay of the memory bus */
  var.dMhsh  = 10;              /* memory delay: reqp, Brdy*/
  for (size=9; size<15; size++)
  { printf("\n DLXp: interrupts, Memory-if: %d; dBUS = %d \n",
          var.BUS,var.dbus);
    printf(" Cache size: 2^ %d \n",size);
 
   for (way=1; way<=4; way = way*2)
   { L1c.way = way;
     L1c.wrpol = 0;
     L1c.boff = 3;
     L1c.sector = 0;
     L1c.line = size - ld(way) - L1c.boff - L1c.sector;
     L1c.tag = var.n - L1c.line - L1c.boff - L1c.sector;
 
     print_line();              /* cost analysis */
     cacheheader("ways",way);
     fL1_print("Menv",cMenv);
     fL1_print("DP",cDP);
     fL1_print("DLX",cDLX);
     print_line();
     fL1_print("aMbusy",aMbusy);
     fL1_print("tstall",tstall);
     fL1_print("tMifC",tMifC);
     print_line();
     fL1_print("tDLX",tDLX);
     fL1_print("T_Mreq",tMreq);
     fL1_print("T_Mrbst",tMrbst);
     fL1_print("T_cache",tcacheIF);
     fL1_print("T_cachRD",tcachRD);
     DRAMfac = 4;  fL1_print("T_M(4)",tM);
     DRAMfac = 8;  fL1_print("T_M(8)",tM);
     print_line();
   }
 }  
}
