/* =========================================================== */
/* pdlxi.c      architectural module   pipelined DLX core      */
/*       with interrupt handling, hardwired control, CLA adder */
/* ----------------------------------------------------------- */
/*  last update: 99/10/13 SMM                                  */
/* =========================================================== */

#include <stdio.h>
#include <math.h>
#include "techno.h"
#include "macro.h"
#include "EXmacro.h"
#include "hardctr.h"
#include "evaluate.h"

struct variable_parameters var;

int stage;
/* ----------------------------------------------------------- */
/* identifies the two control automata                         */
/* stage = 1: Mealy       2: Moore                             */
/* ----------------------------------------------------------- */
/* parameter sets                                              */
/* states, zeta - in, out - monom, lmax, lsum - numax, nusum - */
/* fanmax, fansum - type, binary - aIN, aclr                   */
/* ----------------------------------------------------------- */
struct control_parameters conMealy =
       {1,0, 0,9, 4,10,17, 2,13, 1,0, 2,0, 0,0};
struct control_parameters conMoore = 
       {17,5, 12,16, 26,12,189, 11,48, 3,26, 1,0, 0,0};

/*=============================================================*/
/*   Cost of the data paths and its environments               */
/*=============================================================*/
/*---- IR Environment ----------------------------------------*/
int cIRenv()
{ return 15*cmux +var.n*cff;
}

/*---- CAddr (destination address Cad) -----------------------*/
int cCAddr()
{ return 2* ld(var.A)* cmux;
}

/*---- Daddr (register file addresses) -----------------------*/
int cDaddr()
{ return cCAddr() + 2*5*cmux;
}

/*---- GPR register file -------------------------------------*/
int cGPR()
{ return  cram2(var.A, var.n) 
	  + 2* (czero(ld(var.A)) + cinv + var.n* cand);
}

/*---- Special register file SF ------------------------------*/
int cSF(K,n)
int K,n;
{ return  K*n* (cff + cmux) + 2*cdecf(ld(K)) + cand *K *(n+1) 
	  + (K + n* ctree(K)) *cor;
}

/*---- register file SPR -------------------------------------*/
int cSPR()
{ return  cSF(var.S, var.n) 
	  +3 * var.n * cmux              /* SPRsel */
	  +czero(ld(var.S)) +2*cand +cinv;  /* SPRglue */
}

/*---- execute environment EXenv -----------------------------*/
int cEXenv()
{ return var.n* (2*cmux + 6*cdriv) + cALUenv() + cshenv();
}

/*------------------------------------------------------------*/
/*   Cause environment                                        */
/*------------------------------------------------------------*/
/*---- Cause collection --------------------------------------*/
int cCAcol()
{ return  cand +cor +9*cff;
}

/*---- Cause processing --------------------------------------*/
int cCApro()
{ return  28*cand +cor*(ctree(var.n) + 1) +cinv +cnor +34*cff;
}

/*---- Cause environment--------------------------------------*/
int cCAenv()
{ return  cCApro() + cCAcol();
}

/*-------------------------------------------------------------*/
/*   PC environment (delayed PC)                               */
/*-------------------------------------------------------------*/
/*---- PC glue logic -----------------------------------------*/
int cPCglue()
{ return 2* cor + cand + cxor + czero(var.n);
}

/*---- PC data paths -----------------------------------------*/
int cPCenv()
{ return var.n* (4*cff + 6*cmux) + cCLAf(var.n) 
	  + cincf(var.n-2) + cPCglue();
}

/*---- GPR forwarding data paths (+control, 2x Forw3, +top) --*/
int cGforw()
{ int cost, ldA;

  ldA = ld(var.A);
  cost = ldA*cxor + czero(ldA) + 3*cand + var.n*cmux;

  cost = 3*cost + 3*cand + 2*cinv + ctree(ldA)*cor;
  return 2*cost;
}

/*-------------------------------------------------------------*/
/*   SPR forwarding: general x-stage,  + SR, EPC and EDPC      */
/*-------------------------------------------------------------*/
/*---general x-stage SPR forwarding ---------------------------*/
int csfor(x)
int x;
{ return var.n*x*cmux + x*(cEQ(ld(var.S)) +2*cand);
}

/*--- Forwarding of SR and EPC --------------------------------*/
int cSforw()
{ return csfor(1) +2*csfor(3);
}

/*---- buffering of addresses and PCs ------------------------*/
int cBuffer()
{ return cff *(6* var.n +3*(2*ld(var.A) + 12));
}

/*-------------------------------------------------------------*/
/*   whole DLX data paths                                      */
/*-------------------------------------------------------------*/
int cDP()
{ return cPCenv() + cIRenv() + cEXenv() + cGPR() + cSH4Lie()
	 + cSPR() + cCAenv() + cDaddr() 
	 + cBuffer() + cGforw() + cSforw() + 8*var.n*cff;
}

/*=============================================================*/
/*   Cost of the control: stall, automata, IMC, DMC            */
/*=============================================================*/
int cIMC()
{ return cor;
}

int cDMC()
{ return cdecf(2) + 15*cand + 8*cor + 3*cinv;
}

int cstall()
{ int cdhaz, cdhazS, cCE, cENG;

  cdhazS = 3*cEQ(ld(var.S)) +7*cand +2*cor;
  cdhaz = cor + cnor + 2*(cor + cand + cinv ) + cdhazS;
  cENG = 5*cand + cor + 3*cff;       /* full[] + ue[] */

  /* busy signals, CE1, CE2, JISR */
  cCE = cdhaz + 4*cand +3*cor +4*cnor +cnand;

  return cCE +cENG +cnand + cnor + cor + cinv + 9*cand;
}

/* --- control automata -------------------------------------- */
int cpauto()
{ int cost;

  con = conMealy; stage = 1; cost = cauto();   /* Mealy */
  con = conMoore; stage = 2; cost += cauto();  /* pMoore(EX-WB) */
  return cost;
}

/* ------------------------------------------------------------ */
/* signal of type y/z buffered (xyz.2 stored in automaton)      */
/* valid signals v[4:2] are used for forwarding/stalling        */
/* ------------------------------------------------------------ */
int cCONbuffer()
{ int cvalid;

  cvalid = 6*cff + cinv;      /* v[2:4].i */
  return cvalid + cff*(2*3 + 1);
}


int cCON()
{ return cIMC() +cDMC() +cstall() +cpauto() +cCONbuffer();
}


/*=============================================================*/
/*   Control Automata: Delay of the control signals            */
/*=============================================================*/
int acsMealy()                 /* Mealy signals: ID stage */
{ con = conMealy;
  return aOut();
}

int acsMoore()                  /* Moore signals */
{ con = conMoore;
  return aOut();
}

int aJISR()                     /* JISR = jisr.4 ^ full.4 */
{ return max(dand, dnand);
}

/*=============================================================*/
/*   Delay / Cycle Time of the data paths environments         */
/*=============================================================*/
/* --- execute stage EXenv ----------------------------------- */
int aEX()
{ return acsMoore() +dmux +max(dALUenv(), dALUe2()+dshenv())
	      + ddriv;
}

int aALUovf()
{ return acsMoore() +dmux +dALUovf();
}

int tEX()
{ return aEX() + dff + delta;
}

/* ---------------------------------------------------------- */
/*  WB stage: SH4L GPR (# Registers: var.A, width: var.n)     */
/*                 SPR (# Registers: var.S, width: var.n)     */
/* ---------------------------------------------------------- */
/*---- Special register file SF ------------------------------*/
int dSFdout(K)                /* std. data output */
int K;
{ return  ddecf(ld(K)) + dand +dtree(K) *dor;
}

int dSFw(K)                   /* delay on write */
int K;
{ return  ddecf(ld(K)) + dand +dor +dff;
}

int dSFw_ce()                 /* SFw[],SFw/r -> ce*/
{ return  dand +dor;
}

/*---- register file SPR -------------------------------------*/
int aSPRsin()                   /* delay of std output Sin */
{ return dSFdout(var.S);
}

int tSPR()                   /* write cycle time  of SPR */
{ return max(dzero(ld(var.S)), aJISR()) 
	   +2*dand +dmux +dSFw(var.S) +delta;
}

/*---- register file GPR -------------------------------------*/
int dGPRread()                   
{ return max(dram2(var.A, var.n), dzero(ld(var.A)) +dinv) 
	 + dand;
}

int dGPRwrite()                   
{ return dram2(var.A, var.n);
}

int aGPRin()                  /* input Cdata of GPR */
{ return dSH4Lie();
}

int tWB()                   /* write cycle time  tGPR, tSPR */
{ return max(tSPR(), aGPRin() + dGPRwrite() + delta);
}

/* ---------------------------------------------------------- */
/*   delay of the forwarding engine GPR                       */
/* ---------------------------------------------------------- */
int dforw_Do()                                /* -> Dout */
{ return dxor + 2*dand + 3*dmux + dzero(ld(var.A));
}

int dforw_top()                               /* -> top */
{ return dxor + 4*dand + 3*dmux + dzero(ld(var.A));
}

int dforw_Dat()                         /* data -> Dout */
{ return 3*dmux;
}

int aABin()
{ int aDataIn;

  aDataIn = max(max(aEX(), aGPRin()), dGPRread()) +dforw_Dat();
  return max(aDataIn, dforw_Do());
}

/* ---------------------------------------------------------- */
/*   delay of the forwarding engine SPR                       */
/* ---------------------------------------------------------- */
int dSfor_hit()                               /* -> hit */
{ return dEQ(ld(var.S)) + dand;
}

int dSfor_Do(x)                               /* -> Dout */
int x;
{ return dSfor_hit() + x*dmux;
}

int dSfor_Dat(x)                         /* data -> Dout */
int x;
{ return (x)*dmux;
}

int aSRfor()
{ return max(aJISR() + dmux +dSfor_Dat(1), dSfor_Do(1)); 
}

int aEPCfor()
{ return max(dSfor_Do(3), max(aEX(), aJISR()+dmux) +dSfor_Dat(3));
}

/* ---------------------------------------------------------- */
/*     ID stage: IRenv, CAddr, PCenv, GPRread                 */
/* ---------------------------------------------------------- */
/*---- operand fetch ---------------------------------------- */
int tABread()                              /* read cycle time */
{ return aABin() + dff +delta;
}

/*---- IR Environment (Mealy) ------------------------------- */
int aIRco()
{ return acsMealy() + dmux;
}

int tIRenv()
{ return aIRco() + dff + delta;
}

/*---- Destination Address CAddr ---------------------------- */
int tCAddr()
{ return acsMealy() + 2*dmux + dff + delta;
}

/* --- PC environment (delayed PC) -------------------------- */
int adpc()                   
{ return aJISR() + 2*dmux;      /* EDPC never forwarded */
}

int aPCglue()                   
{ int delay;

  delay = dor + dand + dxor;          /* dPCglue */
  return max(acsMealy(), aABin() +dfzero(var.n)) + delay;
}

int tPCenv()                   
{ int delay;

  delay = max(dincf(var.n -2), aIRco() + dCLAf(var.n));
  delay = max(max(delay, aABin()), max(aPCglue(), aEPCfor()) );
  return delay + 3*dmux + dff + delta;
}

int tID()                   
{ return max(max(tPCenv(), tIRenv()), max(tCAddr(), tABread()));
}

/*=============================================================*/
/*   Delay / Cycle Time of the memory stages: IF, M            */
/*=============================================================*/
int aDMC()                  /* control signals of DMC */
{ int aDMC_MW, delay;

  aDMC_MW = aJISR() + 2*dand + dnor + dor;

  delay = acsMoore() + max(ddecf(2), dinv + dand) + 2*(dand+dor);

  return max(delay, aDMC_MW);
}

int aIMC()                  /* control signals of IMC */
{ return adpc();
}

int tIF()
{ return aIMC() + var.dImem + dff + delta;
}

int aMbusy()                       /* signals ibusy, dbusy */
{ return max(aIMC() +var.dIstat, aDMC() + var.dstat);
}

int tDM()
{ return aDMC() + var.dmem + dff + delta;
}

/* --- Memory cycle time tMM -------------------------------- */
int tMM()
{ return max(tIF(), tDM());
}
 
/*=============================================================*/
/*   Cause environment: collection and processing              */
/*=============================================================*/
int aCAcol()                             /* A_CAcol(CA.4) */
{ return aDMC() +max(dor, var.dstat);
}

int tCAcol()                         /* CAcol: register update */
{ return max(adpc()+var.dIstat, aALUovf()+dand) + dff + delta;
}

int tCApro()                         /* CApro: register update */
{ return max(aSRfor(), aCAcol()) 
	   +dand + dor*dtree(var.n) + dff + delta;
}

int tCAenv() 
{ return max(tCAcol(), tCApro());
}

/*== cycle time: data paths DP ================================*/
int tDP()
{ return max( max(tWB(), tEX()), max(tID(), tCAenv()) );
}
 
/*============================================================*/
/*   Control Environment (including pAuto, Cce, IMC, DMC)     */
/*============================================================*/
int tpauto()
{ int delay;

  stage = 1; con = conMealy; 
  delay = tauto();                      /* Mealy(ID) */

  con = conMoore; stage = 2;            /* pMoore(EX-WB) */
  return max(delay, tauto()); 
}

int tCONbuf()
{ return tpauto() + dand;
}

int tstall()
{ int Anotbusy, AnotDhaz, aCE, Aue;

  Anotbusy = 2*dnor +dand +aMbusy();
  AnotDhaz = dnor + max(dforw_top() +dand +2*dor,
			dand + 2*dor + dSfor_hit(3)); 

  aCE = max(Anotbusy, AnotDhaz) +dand +dor; 
  Aue = aCE + 2*dand;
 
  return Aue +dand +max(dSFw_ce() +dff, dGPRwrite()) +delta; 
}

/* ----- Control Unit --------------------------------------- */
int tCON()
{ return max(tCONbuf(), tstall());
}

/* ---------------------------------------------------------- */
/*     Cost and cycle time of the DLX Design                  */
/* ---------------------------------------------------------- */
int cDLX()
{ return cDP() + cCON();
} 

int tDLX()
{ return max(tDP(), tCON());
} 

int cCONrel()
{ return (int)((100.0* cCON())/ cDLX());
}

int cGPRrel1()
{ return (int)((100.0* cGPR())/ cDP());
}

int cGPRrel2()
{ return (int)((100.0* cGPR())/ cDLX());
}


/* === Print functions ====================================== */
void print_line()
{
  printf(
  "---------------------------------------------------------\n");
  return;
}

void print_header(str1,str2)
char *str1, *str2;
{
  printf("%-14s \t      IDmem / IDstat \n",str1);
  printf("%-14s \t    0/0     25/5    0/50    50/0\n",str2);
  print_line();
  return;
}

void print_value(str,func)
char *str;
int (*func)();
{ printf("%14s:",str); 
  var.dstat = var.dIstat = 0;  var.dmem = var.dImem = 0;  
  printf("\t%7d",(*func)());
  var.dstat = var.dIstat = 5;  var.dmem = var.dImem = 25;  
  printf("\t%7d",(*func)());
  var.dstat = var.dIstat =50;  var.dmem = var.dImem = 0;  
  printf("\t%7d",(*func)());
  var.dstat = var.dIstat = 0;  var.dmem = var.dImem = 50;  
  printf("\t%7d\n",(*func)());
  return;
}


/* ---------------------------------------------------------- */
/*     Evaluation of the design                               */
/* ---------------------------------------------------------- */

void main()
{
 char my_str[200];             /* temporary strings */
 int i, x, y;                  /* loop variables */
 int c1, c2, t1, t2;           /* cost, time values */
 double res, q;                /* result, quality parameter */

  init_techno(MOTOROLA);        /* select technology */

  var.n      = 32;              /* width of the data paths */
  var.A      = 32;              /* number of GP-Registers */
  var.S      = 6;               /* number of SP-Registers */
  var.dmem  = var.dImem = 25;  /* memory access times */
  var.dstat = var.dIstat = 5;  /* memory status times */

  printf("\n\n\n Pipelined DLX, hardwired control, with interrupt\n");
  printf(" Techn.: %s, ",name_techno());
  strcpy(my_str,"CLA"); 
  printf(" %s Adder,", my_str);
  printf(" width n = %d,  Reg-# = %d\n", var.n,var.A);

/* ---------------------------------------------------------- */
/*  Cost of the DP environments and the control               */
/* ---------------------------------------------------------- */

  print_line();              /* cost analysis */
  print_header("cost","");
  print_value("EXenv",cEXenv);
  print_value("SH4Le",cSH4Lie);
  print_value("IRenv",cIRenv);
  print_value("Daddr",cDaddr);
  print_value("GPRenv",cGPR);
  print_value("SPRenv",cSPR);
  print_value("CAenv",cCAenv);
  print_value("PCenv",cPCenv);
  print_value("Sforw",cSforw);
  print_value("Gforw",cGforw);
  print_value("Buffer",cBuffer);
  print_line();
  
  print_value("stall",cstall);
  print_value("MDC",cDMC);
  print_value("pauto",cpauto);
  print_value("CONbuf",cCONbuffer);
  print_line();

  print_value("DP",cDP);
  print_value("Control",cCON);
  print_value("DLX",cDLX);
  print_line();

  print_value("CON/DLX [%]",cCONrel);
  print_value("GPR/DP [%]",cGPRrel1);
  print_value("GPR/DLX [%]",cGPRrel2);
  print_line();
  
/* ---------------------------------------------------------- */
/*    Cycle time and delay of DP environments                 */
/* ---------------------------------------------------------- */
  print_line();              /* cycle time analysis */
  print_header("delay","  ");

  print_value("tID",tID);
  print_value("tPCenv",tPCenv);
  print_value("tABread",tABread);
  print_value("tEX",tEX);
  print_value("tWB",tWB); 
  print_value("tSPR",tSPR); 
  print_value("tDP",tDP);
  print_line();
  print_value("acsMealy",acsMealy);
  print_value("aABin",aABin);
  print_value("aPCglue",aPCglue);
  print_line();
  print_value("aMbusy",aMbusy);
  print_value("tCONbuf",tCONbuf);
  print_value("tstall",tstall);
  print_value("tCON",tCON);
  print_line();
  print_value("tDLX",tDLX);
  print_line();
  print_value("tIF,M",tMM);
  print_line();
  printf("\n\n\n");
}
