/* =========================================================== */
/* DLXs.c       architectural module   non-pipelined DLX       */
/*         no interrupt handling, hardwired control, CLA adder */
/* ----------------------------------------------------------- */
/*  last update: 9812oct SMM                                   */
/* =========================================================== */
/*  BUS =  0: mbusy                                           */
/*         1: 3-line handshake interface                      */
/*         2: direct mapped cache (line, sector, bytes)       */
/* ---------------------------------------------------------- */


#include <stdio.h>
#include <math.h>
#include "techno.h"
#include "macro.h"
#include "EXmacro.h"
#include "hardctr.h"
#include "cache.h"
#include "evaluate.h"

struct variable_parameters  var;

/* Main Control: sequential DLX without L1-cache */
/*------------------------------------------------------- */
/* states, log(states), input, output, monom, lmax, lsum  */
/* numax, nusum, fanmax, fansum, type, binary, aIN, aclr  */
/*------------------------------------------------------- */
struct control_parameters mcon = 
    { 23, 5, 13, 29, 20, 11, 115, 12, 94, 4, 33, 1,0,0,0};   

/* Memory Controller: with 3-line protocol */
struct control_parameters conM = 
    { 3, 2, 2, 2, 3, 1, 3, 2, 3, 2, 3, 2,0,0,0};   

/* Memory Controller: with sectored cache */
/* cost and cycle time */
struct control_parameters conIF1 =
    {9,4,5,15, 13,2,22, 7,40, 3,17, 1,0,0,0};
/* delay of the Mealy signals */
struct control_parameters conIF2 =
    {9,4,5,15, 13,2,22, 4,40, 3,17, 3,0,0,0};

struct cache_parameters L1c;
struct cache_parameters L1cache = { 18, 10, 1, 3, 1, 0};
/*       { 22,  7, 0, 3, 1, 0};        8B  1KB */
/*       { 19, 10, 0, 3, 1, 0};        8B  8KB */
/*       { 19,  9, 1, 3, 1, 0};       16B  8KB */
/*       { 18, 10, 1, 3, 1, 0};       16B 16KB */
/*{tag, line, sector, boff, way, wr-through} */


/* =========================================================== */
/*   Cost of the data paths and its environments               */
/* =========================================================== */
/*   Memory environment Menv                                   */
/* ----------------------------------------------------------- */
/* --- Memory Interface Mif ---------------------------------- */
int cMif()
{ 
  switch(var.BUS)
  { case 0:          /* no bus interace -- mbusy */
      return 0;

    case 1:          /* 3-line bus interace */
      return 2*cdriv * var.n;

    default:         /* sectored cache */
      return 3*cdriv * var.n;
   }
}

/* --- Memory Environment Menv ------------------------------- */
int cMenv()
{ 
  if (var.BUS<2 )
    return cMif();  /* no cache */
  else
    return cMif() + cmux*var.n + CchIF();
}

/* --- IR Environment --------------------------------------- */
int cirenv()
{ return 15*cmux +var.n*cff;
}

/* --- GPR Environment (# Registers: var.A, width: var.n) --- */
int cgpr()
{ return  cram2(var.A, var.n) + 2*czero(ld(var.A)) + 2* cinv 
          + 2* ld(var.A)* cmux + 2* var.n* cand;
}

/* --- PC Environment --------------------------------------- */
int cpcenv()
{ return  var.n*cff + var.n*cmux + cor;
}

/*=============================================================*/
/*   Cost of the control: stall, automata, memory control      */
/*=============================================================*/
/* --- Memory Control circuit MC ----------------------------- */
int cMC()
{ int cost;

  cost = cdecf(2) + 12*cand + 5*cor + 2*cinv;

  if (var.BUS < 2 )
    return cost;  /* no cache */
  else
    return cost + cinv + cache.boff * (cand + cff);
}

/* --- Memory Interface Control MifC ------------------------- */
int cMifC()
{ int cost;

  switch(var.BUS)
  { case 0:          /* no bus interace -- mbusy */
      return 0;

    case 1:          /* 3-line bus interace */
      con = conM;
      return  2*cor + cauto();

    default:         /* sectored cache */
      con = conIF1;
      return  cauto();
   }
}

/* ----- Control Unit --------------------------------------- */
int cstall() 
{ return cinv + cor + 10*cand;
}

int cCON() 
{ int cost;
  
  con = mcon;
  cost = cauto();
  return cost + cMC() + cstall() + cMifC();
}

/*============================================================*/
/*   Delay of the data paths                                  */
/*============================================================*/
/* --- IR Environment --------------------------------------- */
int direnv()
{ return dmux;
}

/* --- GPR Environment (# Registers: var.A, width: var.n) --- */
int dGPRw()
{ return 2*dmux + dram2(var.A, var.n);
}

int dGPRr()
{ return  max(dram2(var.A, var.n), dzero(ld(var.A)) +dinv) 
          + dand;
}

int aAEQZ()
{ return dzero(var.n);
}

/* --- PC Environment --------------------------------------- */
int dpcIN()
{ return max(dor, dmux);
}

/* ----------------------------------------------------------- */
/*     Delay: outputs Csig of the control automaton            */
/* ----------------------------------------------------------- */
int aCON_Csig() 
{ con = mcon;
  return aOut();
}

/* ---------------------------------------------------------- */
/*     Cost and delay of the DLX Data Paths                   */
/* ---------------------------------------------------------- */
int cDP()
{ return cALUenv() + cshenv() + cSH4Le() + cirenv() + cgpr() 
        + cpcenv() +var.n*(6*cff +9*cdriv +cmux) + cMenv();
}

/* --- clocking of registers in the data paths -------------- */
int tRegce()
{ return aCON_Csig() + dand + dff + delta;
}

/* --- register file write back ----------------------------- */
int tGPRw()
{ return aCON_Csig() +dSH4Le() +dGPRw() + delta;
}

/* --- register file read cycle ----------------------------- */
int tGPRr()
{ return aCON_Csig() + dGPRr() + dff + delta;
}

/* --- data on operand busses a,b---------------------------- */
int aBUSab()
{ return aCON_Csig() + direnv() +ddriv;
}

/* --- Cycles through ALU or shifter SH --------------------- */
int tALU_SH()
{ return aBUSab() +max(dshenv(), dALUenv()) + ddriv + dpcIN() 
         + dff + delta;
}
 
/* --- Cycle time of the data paths ------------------------- */
int tDP()
{ return max( max(tGPRw(), tGPRr()), max(tALU_SH(), tRegce() ));
}


/* =========================================================== */
/*   Memory System: delay + cycle times                        */
/* =========================================================== */
/* --- Set Delay of cacheIF inputs --------------------------- */
int INcache()
{ 
  cacheIF.A = aCON_Csig() + dmux;
  cacheIF.Din = cacheIF.csig =0;
}

/* --- Memory Control circuit MC ----------------------------- */
int aMC() 
{ int delay;
  
  delay = max(ddecf(2), dinv + dand) + 2*dand + 2*dor;

  if (var.BUS <2)
    return aCON_Csig() + delay;
  else
    return aCON_Csig() + dand + delay;
}

int aMC_BE() 
{ 
  if (var.BUS <2)
    return aMC();
  else
    return 0;             /* with cache: MBW are buffered */
}

/* --- Memory Interface Control MifC ------------------------- */
int aMifC()
{ int delay;
  
  switch(var.BUS)
  { case 0:                      /* no bus interace -- mbusy */
      return 0;

    case 1:                          /* 3-line bus interace */
      conM.aIN = aCON_Csig() + dor; con = conM;   
      return  aOut();

    default:                     /* sectored cache */ 
      return 0;
  }
}

 
/* --- Memory cycle time tM --------------------------------- */
int tM()                                 /* t_Maccess */
{ int delay;

  switch(var.BUS)
  { case 0:                      /* no bus interace -- mbusy */
      return aMC() +var.dmem +dff +delta;

    case 1:                      /* 3-line bus interace */
      return max(aMC_BE(), aMifC() + ddriv) +var.dbus 
                 +dDRAM(var.Mword,8) +delta;

    default:                      
     INcache();
     delay = DchIFM_ch();
     delay = max(delay, aMifC()) + ddriv;
     return max(delay, aMC_BE()) +var.dbus 
                 +dDRAM(var.Mword,8) +delta;
  }
}

int tcacheIF()                               /* t_cacheIF */
{ int delay;

  if (var.BUS < 2)
    return 0;
  else
  { INcache();
    return  TchIF();
  }
}

int tMrbst()                               /* t_Mrburst */
{ int delay;

  if (var.BUS < 2)
    return 0;
  else
  { 
    INcache();
    delay = max(DchIFM_ch(), DchIFM_Dout()+ dmux +dff);
    return  delay + ddriv +var.dbus +delta;
  }
}

int tMreq()                               /* t_Mrequest */
{ int delay;

  switch(var.BUS)
  { case 0:                      /* no bus interace -- mbusy */
      return 0;

    case 1:                      /* 3-line bus interace */
      return aMifC() + ddriv +var.dbus + var.dMhsh + dff +delta;

    default:                      /* aMifC()=0 */
      INcache();
      delay = AchIF_MAdr();
      return delay + ddriv +var.dbus + var.dMhsh + dff +delta;
  }
}

int tcachRD()                               /* t_$rd */
{ int delay;

  if (var.BUS < 2)
    return 0;
  else
  { 
    INcache();
    return  AchIF_Dout()+ dmux +dff;
  }
}

/* --- Acc. dealy of mbusy ---------------------------------- */
int Ambusy()
{int delay;

  switch(var.BUS)
  { case 0:                      /* no bus interace -- mbusy */
      return aMC() +var.dstat;

    case 1:                      /* 3-line bus interace */
      return aMifC();

    default:                      /* aMifC(mbusy)>0 */
      INcache();
      delay = max(AchIF_hit(), aCON_Csig());
      conIF2.aIN = max(delay, var.dbus);   con = conIF2;
      return aOut();
  }
}
 
/* --- Stalling: mbusy -> Reg, GPR--------------------------- */
int tstall()
{int delay;

 delay = Ambusy() +dor +dand +delta 
	+ max(dff, dram2(var.A, var.n));

 if (var.BUS < 2)
  return delay + dinv;    /* MifC provides mbusy */
 else 
  return delay;           /* MifC provides /mbusy */
}
 
/* ---------------------------------------------------------- */
/*   Control Environment (2): Next State Computation          */
/* ---------------------------------------------------------- */
int tCONmain()      /* main control automaton */
{
  mcon.aIN = aAEQZ();   mcon.aclr = 0;   con = mcon;
  return tauto();
}

/* --- Memory Interface Control MifC ------------------------- */
int tMifC()          /* automaton in MifC */
{ int delay;
  
  switch(var.BUS)
  { case 0:                      /* no bus interace -- mbusy */
      return 0;

    case 1:                          /* 3-line bus interace */
      conM.aIN = max(aCON_Csig() +dor, var.dbus); 
      conM.aclr = 0;   con = conM;
      return  tauto();

    default:                     /* sectored cache */ 
      INcache();
      delay = max(AchIF_hit(), aCON_Csig());
      conIF1.aIN = max(delay, var.dbus);   
      conIF1.aclr = 0;   con = conIF1;
      delay = tauto();           /* update of MifC automaton */
                                 /* buffering of MBW */
      return  max(delay, aMC()+dff+delta);
  }
}

int tCON()
{ int delay;

  delay = max(tCONmain(), tMifC());
  return max(delay, tstall());
}

/* ---------------------------------------------------------- */
/*     Cost and cycle time of the DLX Design                  */
/* ---------------------------------------------------------- */
int cDLX()
{ return cDP() + cCON();
} 

int tDLX()
{ return max(tDP(), tCON());
} 

int cCONrel()
{ return (int)((100.0* cCON())/ cDLX());
}

int cGPRrel1()
{ return (int)((100.0* cgpr())/ cDP());
}

int cGPRrel2()
{ return (int)((100.0* cgpr())/ cDLX());
}

/* === Print functions ====================================== */
void print_line()
{
  printf(
  "---------------------------------------------------------\n");
  return;
}

void print_header(str1,str2)
char *str1, *str2;
{
  printf("%-14s \t      Dmem,bus / Dstat,Mhsh \n",str1);
  printf("%-14s \t    0/0     15/10   0/20    50/0\n",str2);
  print_line();
  return;
}

void print_value(str,func)
char *str;
int (*func)();
{ printf("%14s:",str); 
  var.dstat = var.dIstat = var.dMhsh = 0;
  var.dmem = var.dImem = var.dbus = 0;  

  printf("\t%7d",(*func)());
  var.dstat = var.dIstat = var.dMhsh = 10;
  var.dmem = var.dImem = var.dbus = 15;  

  printf("\t%7d",(*func)());
  var.dstat = var.dIstat =var.dMhsh = 20;
  var.dmem = var.dImem = var.dbus = 0;  

  printf("\t%7d",(*func)());
  var.dstat = var.dIstat = var.dMhsh = 0;
  var.dmem = var.dImem = var.dbus = 50;  
  printf("\t%7d\n",(*func)());
  return;
}

void cacheheader(str1,value)
char *str1;
int value;
{
  printf("%-11s \t      Line size [Bytes] \n",str1);
  printf("%-11d \t    8       16     32       64    128 \n",value);
  forall_line();
  return;
}

void fL1_print(str,func)
char *str;
int (*func)();
{ printf("%11s:",str);
  cache = L1c;  
  printf("\t%7d",(*func)());

  cache.sector +=1; cache.line -=1;     /* 16B*/
  printf("\t%7d",(*func)());

  cache.sector +=1; cache.line -=1;     /* 32B*/
  printf("\t%7d",(*func)());

  cache.sector +=1; cache.line -=1;     /* 64B*/
  printf("\t%7d",(*func)());

  cache.sector +=1; cache.line -=1;     /* 128B*/
  printf("\t%7d\n",(*func)());
  return;
}

/* ---------------------------------------------------------- */
/*     Evaluation of the design                               */
/* ---------------------------------------------------------- */

void main()
{
  int i, way, size;
  char my_str[200];             /* temporary strings */

  init_techno(MOTOROLA);        /* select technology */

  con = mcon;
  cache = L1cache;
                                /* design parameters */
  var.n      = 32;              /* width of the data paths */
  var.A      = 32;              /* number of GP-Registers */
  var.dmem   = 25;              /* memory access time */
  var.dstat = 5;                /* memory status time */
  var.dbus   = 15;              /* delay of the memory bus */
  var.dMhsh  = 10;              /* memory delay: reqp, Brdy*/
  var.Mword  = 16 *1024 * 1024; /* memory size in words */

  for(i=0; i<=2; i++)
  { 
    var.BUS = i;
    cache = L1cache;

    printf("\n DLX Design: no pipelining, no interrupts \n");
    printf(" Memory interface : %d; delay dBUS = %d \n",
          var.BUS,var.dbus);
    printf(
" Cache: lines= %d, sector= %d, offset= %d, way= %d, Write: %d \n",
    cache.line,cache.sector,cache.boff,cache.way,cache.wrpol);

/* ---------------------------------------------------------- */
/*  Cost of the DP environments and the control               */
/* ---------------------------------------------------------- */

    print_line();              /* cost analysis */
    print_header("cost","");
    print_value("ALUenv",cALUenv);
    print_value("SHenv",cshenv);
    print_value("SH4Lenv",cSH4Le);
    print_value("IRenv",cirenv);
    print_value("GPRenv",cgpr);
    print_value("PCenv",cpcenv);
    print_value("Menv",cMenv);
    print_line();
  
    print_value("DP",cDP);
    print_value("Control",cCON);
    print_value("MC",cMC);
    print_value("MifC",cMifC);
    print_value("DLX",cDLX);
    print_line();

    print_value("CON/DLX [%]",cCONrel);
    print_value("GPR/DP [%]",cGPRrel1);
    print_value("GPR/DLX [%]",cGPRrel2);
    print_line();
  
/* ---------------------------------------------------------- */
/*    Cycle time and delay of DP environments                 */
/* ---------------------------------------------------------- */
    print_line();              /* cycle time analysis */
    sprintf(my_str,"(dmem = %d)",var.dmem);
    print_header("delay",my_str);

    print_value("aCON_Csig",aCON_Csig);
    print_value("aBUSab",aBUSab);
    print_value("aAEQZ",aAEQZ);
    INcache();
    print_value("Ahit",AchIF_hit);
    print_value("Ambusy",Ambusy);
    print_line();
    print_value("tGPRwrite",tGPRw);
    print_value("tGPRread",tGPRr);
    print_value("tALU-SH",tALU_SH);
    print_line();
    print_value("tDP",tDP);
    print_value("tCON",tCON);
    print_value("tCONmain",tCONmain);
    print_value("tstall",tstall);
    print_value("tMifC",tMifC);
    print_line();
    print_value("tDLX",tDLX);
    print_value("T_Mreq",tMreq);
    print_value("T_Mrbst",tMrbst);
    print_value("T_cache",tcacheIF);
    print_value("T_cachR",tcachRD);
    DRAMfac = 4;  print_value("T_M(4)",tM);
    DRAMfac = 8;  print_value("T_M(8)",tM);
    DRAMfac = 16; print_value("T_M(16)",tM);
    print_line();
  }

for (i=2; i<3; i++)
{ var.BUS = i;
  var.dbus   = 15;              /* delay of the memory bus */
  var.dMhsh  = 10;              /* memory delay: reqp, Brdy*/
  for (size=10; size<18; size++)
  { printf("\n DLX: no pipelining, Memory-if: %d; dBUS = %d \n",
          var.BUS,var.dbus);
    printf(" Cache size: 2^ %d \n",size);

   for (way=1; way<=4; way = way*2)
   { L1c.way = way;
     L1c.wrpol = 0;
     L1c.boff = 3;
     L1c.sector = 0;
     L1c.line = size - ld(way) - L1c.boff - L1c.sector;
     L1c.tag = var.n - L1c.line - L1c.boff - L1c.sector;

     print_line();              /* cost analysis */
     cacheheader("ways",way);
     fL1_print("Menv",cMenv);
     fL1_print("DP",cDP);
     fL1_print("DLX",cDLX);
     print_line();
    fL1_print("Ambusy",Ambusy);
    fL1_print("tstall",tstall);
    fL1_print("tMifC",tMifC);
    print_line();
    fL1_print("tDLX",tDLX);
    fL1_print("T_Mreq",tMreq);
    fL1_print("T_Mrbst",tMrbst);
    fL1_print("T_cache",tcacheIF);
    fL1_print("T_cachR",tcachRD);
    DRAMfac = 4;  fL1_print("T_M(4)",tM);
    DRAMfac = 8;  fL1_print("T_M(8)",tM);
    print_line();
   }
 }
}

}

